import logging
import os
from copy import deepcopy, copy
from typing import Optional, Sequence, Iterable, Union, Tuple, List
import numpy as np

from pyspark.ml import PipelineModel
from pyspark.sql import SparkSession
from pyspark.sql import functions as F, types as SparkTypes, Window
from tqdm import tqdm

from lightautoml.automl.presets.base import upd_params
from lightautoml.automl.presets.utils import plot_pdp_with_distribution
from lightautoml.dataset.base import RolesDict
from lightautoml.ml_algo.tuning.optuna import OptunaTuner
from lightautoml.pipelines.selection.base import SelectionPipeline, ComposedSelector
from lightautoml.pipelines.selection.importance_based import ModelBasedImportanceEstimator, ImportanceCutoffSelector
from lightautoml.pipelines.selection.permutation_importance_based import NpIterativeFeatureSelector
from lightautoml.reader.tabular_batch_generator import ReadableToDf
from sparklightautoml.automl.blend import SparkWeightedBlender, SparkBestModelSelector
from sparklightautoml.automl.presets.base import SparkAutoMLPreset
from sparklightautoml.automl.presets.utils import (
    calc_feats_permutation_imps,
    replace_dayofweek_in_date,
    replace_month_in_date,
    replace_year_in_date,
)
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.ml_algo.boost_lgbm import SparkBoostLGBM
from sparklightautoml.ml_algo.linear_pyspark import SparkLinearLBFGS
from sparklightautoml.pipelines.features.lgb_pipeline import SparkLGBSimpleFeatures, SparkLGBAdvancedPipeline
from sparklightautoml.pipelines.features.linear_pipeline import SparkLinearFeatures
from sparklightautoml.pipelines.ml.nested_ml_pipe import SparkNestedTabularMLPipeline
from sparklightautoml.pipelines.selection.base import BugFixSelectionPipelineWrapper
from sparklightautoml.pipelines.selection.permutation_importance_based import SparkNpPermutationImportanceEstimator
from sparklightautoml.reader.base import SparkToSparkReader
from sparklightautoml.tasks.base import SparkTask
from sparklightautoml.utils import Cacher, SparkDataFrame

logger = logging.getLogger(__name__)

# Either path/full url, or pyspark.sql.DataFrame
ReadableIntoSparkDf = Union[str, SparkDataFrame]

base_dir = os.path.dirname(__file__)


class SparkTabularAutoML(SparkAutoMLPreset):
    """
    Spark version of :class:`TabularAutoML <lightautoml.automl.presets.tabular_presets.TabularAutoML>`.
    Represent high level entity of spark lightautoml. Use this class to create automl instance.

    Example:

        >>> automl = SparkTabularAutoML(
        >>>     spark=spark,
        >>>     task=SparkTask('binary'),
        >>>     general_params={"use_algos": [["lgb"]]},
        >>>     lgb_params={'use_single_dataset_mode': True},
        >>>     reader_params={"cv": cv, "advanced_roles": False}
        >>> )
        >>> oof_predictions = automl.fit_predict(
        >>>     train_data,
        >>>     roles=roles
        >>> )
    """

    _default_config_path = "tabular_config.yml"

    # set initial runtime rate guess for first level models
    _time_scores = {
        "lgb": 1,
        "lgb_tuned": 3,
        "linear_l2": 0.7,
        "cb": 2,
        "cb_tuned": 6,
    }

    def __init__(
        self,
        spark: SparkSession,
        task: SparkTask,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        nested_cv_params: Optional[dict] = None,
        tuning_params: Optional[dict] = None,
        selection_params: Optional[dict] = None,
        lgb_params: Optional[dict] = None,
        cb_params: Optional[dict] = None,
        linear_l2_params: Optional[dict] = None,
        gbm_pipeline_params: Optional[dict] = None,
        linear_pipeline_params: Optional[dict] = None,
    ):
        if config_path is None:
            config_path = os.path.join(base_dir, self._default_config_path)
        super().__init__(task, timeout, memory_limit, cpu_limit, gpu_ids, timing_params, config_path)

        self._cacher_key = "main_cache"

        self._spark = spark
        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "nested_cv_params",
                "tuning_params",
                "selection_params",
                "lgb_params",
                "cb_params",
                "linear_l2_params",
                "gbm_pipeline_params",
                "linear_pipeline_params",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                nested_cv_params,
                tuning_params,
                selection_params,
                lgb_params,
                cb_params,
                linear_l2_params,
                gbm_pipeline_params,
                linear_pipeline_params,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, train_data: SparkDataFrame, multilevel_avail: bool = False):
        # infer optuna tuning iteration based on dataframe len
        if self.tuning_params["max_tuning_iter"] == "auto":
            if not train_data.is_cached:
                self.tuning_params["max_tuning_iter"] = 5

            length = train_data.count()

            if length < 10000:
                self.tuning_params["max_tuning_iter"] = 100
            elif length < 30000:
                self.tuning_params["max_tuning_iter"] = 50
            elif length < 100000:
                self.tuning_params["max_tuning_iter"] = 10
            else:
                self.tuning_params["max_tuning_iter"] = 5

        if self.general_params["use_algos"] == "auto":
            self.general_params["use_algos"] = [["lgb", "lgb_tuned", "linear_l2", "cb", "cb_tuned"]]
            if self.task.name == "multiclass" and multilevel_avail:
                self.general_params["use_algos"].append(["linear_l2", "lgb"])

        if not self.general_params["nested_cv"]:
            self.nested_cv_params["cv"] = 1

    def get_time_score(self, n_level: int, model_type: str, nested: Optional[bool] = None):

        if nested is None:
            nested = self.general_params["nested_cv"]

        score = self._time_scores[model_type]

        mult = 1
        if nested:
            if self.nested_cv_params["n_folds"] is not None:
                mult = self.nested_cv_params["n_folds"]
            else:
                mult = self.nested_cv_params["cv"]

        if n_level > 1:
            mult *= 0.8 if self.general_params["skip_conn"] else 0.1

        score = score * mult

        # lower score for catboost on gpu
        if model_type in ["cb", "cb_tuned"] and self.cb_params["default_params"]["task_type"] == "GPU":
            score *= 0.5
        return score

    def get_selector(self, cacher_key: str, n_level: Optional[int] = 1) -> SelectionPipeline:
        selection_params = self.selection_params
        # lgb_params
        lgb_params = deepcopy(self.lgb_params)
        lgb_params["default_params"] = {
            **lgb_params["default_params"],
            **{"featureFraction": 1},
        }

        mode = selection_params["mode"]

        # create pre selection based on mode
        pre_selector = None
        if mode > 0:
            # if we need selector - define model
            # timer will be useful to estimate time for next gbm runs
            time_score = self.get_time_score(n_level, "lgb", False)

            sel_timer_0 = self.timer.get_task_timer("lgb", time_score)
            selection_feats = SparkLGBSimpleFeatures(cacher_key=cacher_key)

            selection_gbm = SparkBoostLGBM(cacher_key=cacher_key, timer=sel_timer_0, **lgb_params)
            selection_gbm.set_prefix("Selector")

            if selection_params["importance_type"] == "permutation":
                importance = SparkNpPermutationImportanceEstimator()
            else:
                importance = ModelBasedImportanceEstimator()

            pre_selector = ImportanceCutoffSelector(
                selection_feats,
                selection_gbm,
                importance,
                cutoff=selection_params["cutoff"],
                fit_on_holdout=selection_params["fit_on_holdout"],
            )
            if mode == 2:
                time_score = self.get_time_score(n_level, "lgb", False)

                sel_timer_1 = self.timer.get_task_timer("lgb", time_score)
                selection_feats = SparkLGBSimpleFeatures(cacher_key=cacher_key)
                selection_gbm = SparkBoostLGBM(cacher_key=cacher_key, timer=sel_timer_1, **lgb_params)
                selection_gbm.set_prefix("Selector")

                importance = SparkNpPermutationImportanceEstimator()

                extra_selector = NpIterativeFeatureSelector(
                    selection_feats,
                    selection_gbm,
                    importance,
                    feature_group_size=selection_params["feature_group_size"],
                    max_features_cnt_in_result=selection_params["max_features_cnt_in_result"],
                )

                pre_selector = ComposedSelector([pre_selector, extra_selector])

        return BugFixSelectionPipelineWrapper(pre_selector)

    def get_linear(
        self, cacher_key: str, n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None
    ) -> SparkNestedTabularMLPipeline:

        # linear model with l2
        time_score = self.get_time_score(n_level, "linear_l2")
        linear_l2_timer = self.timer.get_task_timer("reg_l2", time_score)
        linear_l2_model = SparkLinearLBFGS(cacher_key=cacher_key, timer=linear_l2_timer, **self.linear_l2_params)
        linear_l2_feats = SparkLinearFeatures(
            output_categories=True, cacher_key=self._cacher_key, **self.linear_pipeline_params
        )

        linear_l2_pipe = SparkNestedTabularMLPipeline(
            self._cacher_key,
            [linear_l2_model],
            force_calc=True,
            pre_selection=pre_selector,
            features_pipeline=linear_l2_feats,
            **self.nested_cv_params,
        )
        return linear_l2_pipe

    def get_gbms(
        self,
        keys: Sequence[str],
        n_level: int = 1,
        pre_selector: Optional[SelectionPipeline] = None,
    ):

        gbm_feats = SparkLGBAdvancedPipeline(cacher_key=self._cacher_key, **self.gbm_pipeline_params)

        ml_algos = []
        force_calc = []
        for key, force in zip(keys, [True, False, False, False]):
            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            gbm_timer = self.timer.get_task_timer(algo_key, time_score)
            if algo_key == "lgb":
                gbm_model = SparkBoostLGBM(cacher_key=self._cacher_key, timer=gbm_timer, **self.lgb_params)
            elif algo_key == "cb":
                raise NotImplementedError("Not supported yet")
            else:
                raise ValueError("Wrong algo key")

            if tuned:
                gbm_model.set_prefix("Tuned")
                gbm_tuner = OptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                )
                gbm_model = (gbm_model, gbm_tuner)
            ml_algos.append(gbm_model)
            force_calc.append(force)

        gbm_pipe = SparkNestedTabularMLPipeline(
            self._cacher_key,
            ml_algos,
            force_calc,
            pre_selection=pre_selector,
            features_pipeline=gbm_feats,
            **self.nested_cv_params,
        )

        return gbm_pipe

    def create_automl(self, **fit_args):
        """Create basic automl instance.

        Args:
            **fit_args: Contain all information needed for creating automl.

        """
        train_data = fit_args["train_data"]
        multilevel_avail = fit_args["valid_data"] is None and fit_args["cv_iter"] is None

        self.infer_auto_params(train_data, multilevel_avail)
        reader = SparkToSparkReader(cacher_key=self._cacher_key, task=self.task, **self.reader_params)

        pre_selector = self.get_selector(cacher_key="selector_cache")

        levels = []

        for n, names in enumerate(self.general_params["use_algos"]):
            lvl = []
            # regs
            if "linear_l2" in names:
                selector = None
                if "linear_l2" in self.selection_params["select_algos"] and (
                    self.general_params["skip_conn"] or n == 0
                ):
                    selector = pre_selector
                lvl.append(self.get_linear(self._cacher_key, n + 1, selector))

            gbm_models = [
                x for x in ["lgb", "lgb_tuned", "cb", "cb_tuned"] if x in names and x.split("_")[0] in self.task.losses
            ]

            if len(gbm_models) > 0:
                selector = None
                if "gbm" in self.selection_params["select_algos"] and (self.general_params["skip_conn"] or n == 0):
                    selector = pre_selector
                lvl.append(self.get_gbms(gbm_models, n + 1, selector))

            levels.append(lvl)

        # blend everything
        blender = SparkWeightedBlender(max_nonzero_coef=self.general_params["weighted_blender_max_nonzero_coef"])

        # initialize
        self._initialize(
            reader,
            levels,
            skip_conn=self.general_params["skip_conn"],
            blender=blender,
            return_all_predictions=self.general_params["return_all_predictions"],
            timer=self.timer,
        )

    def _get_read_csv_params(self):
        try:
            cols_to_read = self.reader.used_features
            numeric_dtypes = {
                x: self.reader.roles[x].dtype for x in self.reader.roles if self.reader.roles[x].name == "Numeric"
            }
        except AttributeError:
            cols_to_read = []
            numeric_dtypes = {}
        # cols_to_read is empty if reader is not fitted
        if len(cols_to_read) == 0:
            cols_to_read = None

        read_csv_params = copy(self.read_csv_params)
        read_csv_params = {
            **read_csv_params,
            **{"usecols": cols_to_read, "dtype": numeric_dtypes},
        }

        return read_csv_params

    def fit_predict(
        self,
        train_data: ReadableIntoSparkDf,
        roles: Optional[dict] = None,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[ReadableIntoSparkDf] = None,
        valid_features: Optional[Sequence[str]] = None,
        log_file: str = None,
        verbose: int = 0,
    ) -> SparkDataset:
        """Fit and get prediction on validation dataset.

        Almost same as :meth:`lightautoml.automl.base.AutoML.fit_predict`.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names, if can't
              be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example,
              :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features
              if cannot be inferred from `valid_data`.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;
            log_file: Filename for writing logging messages. If log_file is specified,
            the messages will be saved in a the file. If the file exists, it will be overwritten.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        # roles may be none in case of train data is set {'data': np.ndarray, 'target': np.ndarray ...}
        self.set_logfile(log_file)

        if roles is None:
            roles = {}
        read_csv_params = self._get_read_csv_params()
        train, upd_roles = self._read_data(train_data, train_features, read_csv_params)
        if upd_roles:
            roles = {**roles, **upd_roles}
        if valid_data is not None:
            valid_data, _ = self._read_data(valid_data, valid_features, self.read_csv_params)

        oof_pred = super().fit_predict(train, roles=roles, cv_iter=cv_iter, valid_data=valid_data, verbose=verbose)

        return oof_pred

    def predict(
        self,
        data: ReadableIntoSparkDf,
        features_names: Optional[Sequence[str]] = None,
        return_all_predictions: Optional[bool] = None,
        add_reader_attrs: bool = False,
    ) -> SparkDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Parallel inference - you can pass ``n_jobs`` to speedup
        prediction (requires more RAM).
        Batch_inference - you can pass ``batch_size``
        to decrease RAM usage (may be longer).

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
              if cannot be inferred from `train_data`.
            batch_size: Batch size or ``None``.
            n_jobs: Number of jobs.
            return_all_predictions: if True,
              returns all model predictions from last level

        Returns:
            Dataset with predictions.

        """

        read_csv_params = self._get_read_csv_params()

        data, _ = self._read_data(data, features_names, read_csv_params)
        pred = super().predict(data, features_names, return_all_predictions, add_reader_attrs)
        return pred

    def release_cache(self):
        Cacher.release_cache_by_key(self._cacher_key)

    def _read_data(
        self,
        data: ReadableIntoSparkDf,
        features_names: Optional[Sequence[str]] = None,
        read_csv_params: Optional[dict] = None,
    ) -> Tuple[SparkDataFrame, Optional[RolesDict]]:
        """Get :class:`~pyspark.sql.DataFrame` from different data formats.

        Note:
            Supported now data formats:

                - Path to ``.csv``, ``.parquet``, ``.feather`` files.
                - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
                  For example, ``{'data': X...}``. In this case,
                  roles are optional, but `train_features`
                  and `valid_features` required.
                - :class:`pandas.DataFrame`.

        Args:
            data: Readable to DataFrame data.
            features_names: Optional features names if ``numpy.ndarray``.
            n_jobs: Number of processes to read file.
            read_csv_params: Params to read csv file.

        Returns:
            Tuple with read data and new roles mapping.

        """
        if read_csv_params is None:
            read_csv_params = {}

        if isinstance(data, SparkDataFrame):
            return data, None

        if isinstance(data, str):
            path: str = data
            if path.endswith(".parquet"):
                return self._spark.read.parquet(path), None
                # return pd.read_parquet(data, columns=read_csv_params["usecols"]), None
            if path.endswith(".csv"):
                return self._spark.read.csv(path, **read_csv_params), None
            else:
                raise ValueError(f"Unsupported data format: {os.path.splitext(path)[1]}")

        raise ValueError("Input data format is not supported")

    def get_feature_scores(
        self,
        calc_method: str = "fast",
        data: Optional[ReadableIntoSparkDf] = None,
        features_names: Optional[Sequence[str]] = None,
        silent: bool = True,
    ):
        if calc_method == "fast":
            for level in self.levels:
                for pipe in level:
                    fi = pipe.pre_selection.get_features_score()
                    if fi is not None:
                        used_feats = set(self.collect_used_feats())
                        fi = fi.reset_index()
                        fi.columns = ["Feature", "Importance"]
                        return fi[fi["Feature"].map(lambda x: x in used_feats)]

            else:
                if not silent:
                    logger.info2("No feature importances to show. Please use another calculation method")
                return None

        if calc_method != "accurate":
            if not silent:
                logger.info2(
                    "Unknown calc_method. "
                    + "Currently supported methods for feature importances calculation are 'fast' and 'accurate'."
                )
            return None

        if data is None:
            if not silent:
                logger.info2("Data parameter is not setup for accurate calculation method. Aborting...")
            return None

        read_csv_params = self._get_read_csv_params()
        data, _ = self._read_data(data, features_names, read_csv_params)
        used_feats = self.collect_used_feats()
        fi = calc_feats_permutation_imps(
            self,
            used_feats,
            data,
            self.task.get_dataset_metric(),
            silent=silent,
        )
        return fi

    @staticmethod
    def get_histogram(data: SparkDataFrame, column: str, n_bins: int) -> Tuple[List, np.ndarray]:
        assert n_bins >= 2, "n_bins must be equal 2 or more"
        bin_edges, counts = (
            data.select(F.col(column).cast("double"))
            .where(F.col(column).isNotNull())
            .rdd.map(lambda x: x[0])
            .histogram(n_bins)
        )
        bin_edges = np.array(bin_edges)
        return counts, bin_edges

    @staticmethod
    def get_pdp_data_numeric_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: PipelineModel,
        prediction_col: str,
        n_bins: int,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input numeric column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            n_bins (int): The number of bins to produce. Raises exception if n_bins < 2.
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions. Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        counts, bin_edges = SparkTabularAutoML.get_histogram(df, feature_name, n_bins)
        grid = (bin_edges[:-1] + bin_edges[1:]) / 2
        ys = []
        sample_df = (
            df.select(*[c for c in df.columns if c != feature_name])
            .sample(fraction=ice_fraction, seed=ice_fraction_seed)
            .cache()
        )
        for i in tqdm(grid):
            # replace feature column values with constant
            sdf = sample_df.select("*", F.lit(i).alias(feature_name))

            # infer via transformer
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)

        sample_df.unpersist()

        return grid, ys, counts

    @staticmethod
    def get_pdp_data_categorical_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: PipelineModel,
        prediction_col: str,
        n_top_cats: int,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input categorical column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            n_top_cats (int): param to selection top n categories
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions. Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        feature_cnt = (
            df.where(F.col(feature_name).isNotNull()).groupBy(feature_name).count().orderBy(F.desc("count")).collect()
        )
        grid = [row[feature_name] for row in feature_cnt[:n_top_cats]]
        counts = [row["count"] for row in feature_cnt[:n_top_cats]]
        ys = []
        sample_df = (
            df.select(*[c for c in df.columns if c != feature_name])
            .sample(fraction=ice_fraction, seed=ice_fraction_seed)
            .cache()
        )
        for i in tqdm(grid):
            sdf = sample_df.select("*", F.lit(i).alias(feature_name))
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)
        if len(feature_cnt) > n_top_cats:

            # unique other categories
            unique_other_categories = [row[feature_name] for row in feature_cnt[n_top_cats:]]

            # get non-top categories, natural distributions is important here
            w = Window().orderBy(F.lit("A"))  # window without sorting
            other_categories_collection = (
                df.select(feature_name)
                .filter(F.col(feature_name).isin(unique_other_categories))
                .select(F.row_number().over(w).alias("row_num"), feature_name)
                .collect()
            )

            # dict with key=%row number% and value=%category%
            other_categories_dict = {x["row_num"]: x[feature_name] for x in other_categories_collection}
            max_row_num = len(other_categories_collection)

            def get_category_by_row_num(row_num):
                remainder = row_num % max_row_num
                if remainder == 0:
                    key = max_row_num
                else:
                    key = remainder
                return other_categories_dict[key]

            get_category_udf = F.udf(get_category_by_row_num, SparkTypes.StringType())

            # add row number to main dataframe and exclude feature_name column
            sdf = sample_df.select("*", F.row_number().over(w).alias("row_num"))

            all_columns_except_row_num = [f for f in sdf.columns if f != "row_num"]
            feature_col = get_category_udf(F.col("row_num")).alias(feature_name)
            # exclude row number from dataframe
            # and add back feature_name column filled with other categories same distribution
            sdf = sdf.select(*all_columns_except_row_num, feature_col)

            preds = model.transform(sdf)
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)

            grid.append("<OTHER>")
            ys.append(preds)
            counts.append(sum([row["count"] for row in feature_cnt[n_top_cats:]]))

        sample_df.unpersist()

        return grid, ys, counts

    @staticmethod
    def get_pdp_data_datetime_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: PipelineModel,
        prediction_col: str,
        datetime_level: str,
        reader,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input datetime column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            datetime_level (str): Unit of time that will be modified to calculate dependence: "year", "month" or "dayofweek"
            reader (_type_): Automl reader to transform input dataframe before `model` inferring.
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions. Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        df = reader.read(df).data
        if datetime_level == "year":
            feature_cnt = df.groupBy(F.year(feature_name).alias("year")).count().orderBy(F.asc("year")).collect()
            grid = [x["year"] for x in feature_cnt]
            counts = [row["count"] for row in feature_cnt]
            replace_date_element_udf = F.udf(replace_year_in_date, SparkTypes.DateType())
        elif datetime_level == "month":
            feature_cnt = df.groupBy(F.month(feature_name).alias("month")).count().orderBy(F.asc("month")).collect()
            grid = np.arange(1, 13)
            grid = grid.tolist()
            counts = [0] * 12
            for row in feature_cnt:
                counts[row["month"] - 1] = row["count"]
            replace_date_element_udf = F.udf(replace_month_in_date, SparkTypes.DateType())
        else:
            feature_cnt = (
                df.groupBy(F.dayofweek(feature_name).alias("dayofweek")).count().orderBy(F.asc("dayofweek")).collect()
            )
            grid = np.arange(7)
            grid = grid.tolist()
            counts = [0] * 7
            for row in feature_cnt:
                counts[row["dayofweek"] - 1] = row["count"]
            replace_date_element_udf = F.udf(replace_dayofweek_in_date, SparkTypes.DateType())
        ys = []
        sample_df = df.sample(fraction=ice_fraction, seed=ice_fraction_seed).cache()
        for i in tqdm(grid):
            feature_col = replace_date_element_udf(F.col(feature_name), F.lit(i)).alias(feature_name)
            sdf = sample_df.select(*[c for c in sample_df.columns if c != feature_name], feature_col)
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)

        return grid, ys, counts

    def get_individual_pdp(
        self,
        test_data: SparkDataFrame,
        feature_name: str,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ):
        assert feature_name in self.reader._roles
        assert datetime_level in ["year", "month", "dayofweek"]
        assert ice_fraction > 0 and ice_fraction <= 1.0

        pipeline_model = self.make_transformer()

        # Numerical features
        if self.reader._roles[feature_name].name == "Numeric":
            return self.get_pdp_data_numeric_feature(
                test_data, feature_name, pipeline_model, "prediction", n_bins, ice_fraction, ice_fraction_seed
            )
        # Categorical features
        elif self.reader._roles[feature_name].name == "Category":
            return self.get_pdp_data_categorical_feature(
                test_data, feature_name, pipeline_model, "prediction", top_n_categories, ice_fraction, ice_fraction_seed
            )
        # Datetime Features
        elif self.reader._roles[feature_name].name == "Datetime":
            return self.get_pdp_data_datetime_feature(
                test_data,
                feature_name,
                pipeline_model,
                "prediction",
                datetime_level,
                self.reader,
                ice_fraction,
                ice_fraction_seed,
            )
        else:
            raise ValueError("Supported only Numeric, Category or Datetime feature")

    def plot_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        individual: Optional[bool] = False,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        top_n_classes: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ):
        grid, ys, counts = self.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=n_bins,
            top_n_categories=top_n_categories,
            datetime_level=datetime_level,
            ice_fraction=ice_fraction,
            ice_fraction_seed=ice_fraction_seed,
        )

        HISTOGRAM_DATA_ROWS_LIMIT = 2000
        rows_count = test_data.count()
        if rows_count > HISTOGRAM_DATA_ROWS_LIMIT:
            fraction = HISTOGRAM_DATA_ROWS_LIMIT / rows_count
            test_data = test_data.sample(fraction=fraction)
        if self.reader._roles[feature_name].name == "Numeric":
            test_data = test_data.select(F.col(feature_name).cast("double")).toPandas()
        else:
            test_data = test_data.select(feature_name).toPandas()
        plot_pdp_with_distribution(
            test_data,
            grid,
            ys,
            counts,
            self.reader,
            feature_name,
            individual,
            top_n_classes,
            datetime_level,
        )



"""Classes for report generation and add-ons."""

import logging
import math
import os
import warnings
from operator import itemgetter
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyspark.sql.functions as F
import seaborn as sns
from jinja2 import Environment
from jinja2 import FileSystemLoader
from json2html import json2html
from pyspark import RDD
from pyspark.ml.functions import vector_to_array
from pyspark.mllib.evaluation import BinaryClassificationMetrics, RegressionMetrics, MulticlassMetrics
from pyspark.mllib.linalg import DenseMatrix
from pyspark.sql import SparkSession, Column
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.pandas.functions import pandas_udf

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.report.handy_spark_utils import call2
from sparklightautoml.transformers.scala_wrappers.laml_string_indexer import LAMLStringIndexer, LAMLStringIndexerModel

logger = logging.getLogger(__name__)

base_dir = os.path.dirname(__file__)


def extract_params(input_struct):
    params = dict()
    iterator = input_struct if isinstance(input_struct, dict) else input_struct.__dict__
    for key in iterator:
        if key.startswith(("_", "autonlp_params")):
            continue
        value = iterator[key]
        if type(value) in [bool, int, float, str]:
            params[key] = value
        elif value is None:
            params[key] = None
        elif hasattr(value, "__dict__") or isinstance(value, dict):
            params[key] = extract_params(value)
        else:
            params[key] = str(type(value))
    return params


def get_data_for_roc_and_pr_curve(input_data, scores_col_name, true_labels_col_name):
    data = round_score_col(
        input_data=input_data,
        min_co=0.000,
        max_co=1.000,
        step=0.001,
        scores_col_name=scores_col_name,
        true_labels_col_name=true_labels_col_name,
    )

    metrics = BinaryClassificationMetrics(
        scoreAndLabels=data.select(
            F.col(f"{scores_col_name}_rounded").astype("double"), F.col(true_labels_col_name).astype("double")
        ).rdd
    )

    thresholds = metrics.call("thresholds").collect()
    roc = call2(metrics, "roc").collect()[1:-1]
    pr = call2(metrics, "pr").collect()[1:-1]

    df = pd.DataFrame(
        list(zip(thresholds, map(itemgetter(0), roc), map(itemgetter(1), roc), map(itemgetter(1), pr)))
        + [(0.0, 1.0, 1.0, 0.0)]
    )
    df.columns = ["thresholds", "fpr", "recall", "precision"]

    auc_score = metrics.areaUnderROC
    ap_score = metrics.areaUnderPR

    return [df, auc_score, ap_score]


def plot_curves(input_data, scores_col_name, positive_rate, true_labels_col_name, roc_path, pr_path):
    rounded_data, auc_score, ap_score = get_data_for_roc_and_pr_curve(input_data, scores_col_name, true_labels_col_name)

    plot_roc_curve_image(rounded_data, auc_score, roc_path)
    plot_pr_curve_image(rounded_data, ap_score, positive_rate, pr_path)
    return auc_score, ap_score, rounded_data


def round_score_col(input_data, scores_col_name, true_labels_col_name, min_co=0.01, max_co=0.99, step=0.01):

    scores = F.col(scores_col_name)
    _id_col = F.col(SparkDataset.ID_COLUMN)
    true_labels = F.col(true_labels_col_name)

    return input_data.select(
        _id_col, true_labels, (F.ceil(scores / step) * step).alias("_scores_prepared_value")
    ).select(
        _id_col,
        true_labels,
        F.when(F.col("_scores_prepared_value") < min_co, min_co)
        .otherwise(F.when(F.col("_scores_prepared_value") > max_co, max_co).otherwise(F.col("_scores_prepared_value")))
        .alias(f"{scores_col_name}_rounded"),
    )


def plot_roc_curve_image(data, auc_score, path):

    sns.set(style="whitegrid", font_scale=1.5)
    plt.figure(figsize=(10, 10))

    fpr = data["fpr"]
    tpr = data["recall"]

    lw = 2
    plt.plot(fpr, tpr, color="blue", lw=lw, label="Trained model")
    plt.plot([0, 1], [0, 1], color="red", lw=lw, linestyle="--", label="Random model")
    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    lgd = plt.legend(bbox_to_anchor=(0.5, -0.15), loc="upper center", ncol=2)
    plt.xticks(np.arange(0, 1.01, 0.05), rotation=45)
    plt.yticks(np.arange(0, 1.01, 0.05))
    plt.grid(color="gray", linestyle="-", linewidth=1)
    plt.title("Approx ROC curve (GINI = {:.3f})".format(2 * auc_score - 1))
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches="tight")
    plt.close()


def plot_pr_curve_image(data, ap_score, positive_rate, path):
    sns.set(style="whitegrid", font_scale=1.5)
    plt.figure(figsize=(10, 10))

    precision = data["precision"]
    recall = data["recall"]

    lw = 2
    plt.plot(recall, precision, color="blue", lw=lw, label="Trained model")
    plt.plot(
        [0, 1],
        [positive_rate, positive_rate],
        color="red",
        lw=lw,
        linestyle="--",
        label="Random model",
    )
    plt.xlim([-0.05, 1.05])
    plt.ylim([0.45, 1.05])
    plt.xlabel("Recall")
    plt.ylabel("Precision")
    lgd = plt.legend(bbox_to_anchor=(0.5, -0.15), loc="upper center", ncol=2)
    plt.xticks(np.arange(0, 1.01, 0.05), rotation=45)
    plt.yticks(np.arange(0, 1.01, 0.05))
    plt.grid(color="gray", linestyle="-", linewidth=1)
    plt.title("Approx PR curve (AP = {:.3f})".format(ap_score))
    plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches="tight")
    plt.close()


def plot_preds_distribution_by_bins(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))

    box_plot_data = []
    labels = []
    for name, group in data.groupby("bin"):
        labels.append(name)
        box_plot_data.append(group["y_pred"].values)

    box = axs.boxplot(box_plot_data, patch_artist=True, labels=labels)
    for patch in box["boxes"]:
        patch.set_facecolor("green")
    axs.set_yscale("log")
    axs.set_xlabel("Bin number")
    axs.set_ylabel("Prediction")
    axs.set_title("Distribution of object predictions by bin")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_distribution_of_logits(input_data, path, scores_col_name, true_labels_col_name):

    prep_data = round_score_col(
        input_data=input_data,
        min_co=0.000,
        max_co=1.000,
        step=0.001,
        scores_col_name=scores_col_name,
        true_labels_col_name=true_labels_col_name,
    )

    logits_col_name = f"{scores_col_name}_rounded"
    logits_col = F.col(logits_col_name)
    true_labels_col = F.col(true_labels_col_name)

    data = prep_data.select(true_labels_col, logits_col).groupby(true_labels_col, logits_col).count().toPandas()

    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, 10))

    data["proba_logit"] = np.log(data[logits_col_name].values / (1 - data[logits_col_name].values))

    data_0 = data[data[true_labels_col_name] == 0]
    sns.kdeplot(data_0["proba_logit"], shade=True, color="r", label="Class 0 logits", ax=axs, weights=data_0["count"])

    data_1 = data[data[true_labels_col_name] == 1]
    sns.kdeplot(data_1["proba_logit"], shade=True, color="g", label="Class 1 logits", ax=axs, weights=data_1["count"])
    axs.set_xlabel("Logits")
    axs.set_ylabel("Density")
    axs.set_title("Logits distribution of object predictions (by classes)")
    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_pie_f1_metric(data: RDD, path):
    metrics = MulticlassMetrics(predictionAndLabels=data)
    tn, fp, fn, tp = metrics.confusionMatrix().values
    F1 = metrics.fMeasure(1.0)
    prec = metrics.precision(1.0)
    rec = metrics.recall(1.0)

    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(20, 10), subplot_kw=dict(aspect="equal"))

    recipe = [
        "{} True Positives".format(tp),
        "{} False Positives".format(fp),
        "{} False Negatives".format(fn),
        "{} True Negatives".format(tn),
    ]

    wedges, texts = ax.pie([tp, fp, fn, tn], wedgeprops=dict(width=0.5), startangle=-40)

    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
    kw = dict(
        arrowprops=dict(arrowstyle="-", color="k"),
        bbox=bbox_props,
        zorder=0,
        va="center",
    )

    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        ax.annotate(
            recipe[i], xy=(x, y), xytext=(1.35 * np.sign(x), 1.4 * y), horizontalalignment=horizontalalignment, **kw
        )

    ax.set_title(
        "Trained model: Precision = {:.2f}%, Recall = {:.2f}%, F1-Score = {:.2f}%".format(
            prec * 100, rec * 100, F1 * 100
        )
    )
    plt.savefig(path, bbox_inches="tight")
    plt.close()
    return prec, rec, F1


def f1_score_w_co(input_data, true_labels_col_name, scores_col_name, min_co=0.01, max_co=0.99, step=0.01):

    true_labels = F.col(true_labels_col_name)
    scores = F.col(scores_col_name)
    rounded_scores_col_name = f"{scores_col_name}_rounded"
    rounded_scores = F.col(rounded_scores_col_name)

    data = round_score_col(
        input_data=input_data,
        min_co=min_co,
        max_co=max_co,
        step=step,
        true_labels_col_name=true_labels_col_name,
        scores_col_name=scores_col_name,
    )

    _grp = (
        data.groupby(rounded_scores)
        .agg(F.sum(true_labels).alias("sum"), F.count(true_labels).alias("count"))
        .toPandas()
    )
    pos = _grp["sum"].sum()
    neg = _grp["count"].sum() - pos

    positive_rate = pos / _grp["count"].sum()

    grp = _grp.groupby(rounded_scores_col_name).agg(sum=("sum", "sum"), count=("count", "sum"))
    grp.sort_index(inplace=True)

    grp["fp"] = grp["sum"].cumsum()
    grp["tp"] = pos - grp["fp"]
    grp["tn"] = (grp["count"] - grp["sum"]).cumsum()
    grp["fn"] = neg - grp["tn"]

    grp["pr"] = grp["tp"] / (grp["tp"] + grp["fp"])
    grp["rec"] = grp["tp"] / (grp["tp"] + grp["fn"])

    grp["f1_score"] = 2 * (grp["pr"] * grp["rec"]) / (grp["pr"] + grp["rec"])

    best_score = grp["f1_score"].max()
    best_co = grp.index.values[grp["f1_score"] == best_score].mean()

    return best_score, best_co, positive_rate


def get_bins_table(data: DataFrame, n_bins=20):
    df: pd.DataFrame = (
        data.groupby(F.col("y_pred"), F.col("y_true"))
        .count()
        .toPandas()
        .sort_values(by=["y_pred", "y_true"], ascending=[False, False])
    )

    total_count = df["count"].sum()
    bin_size = math.ceil(total_count / n_bins)

    bins_table = data.groupby("bin").agg({"y_true": [len, np.mean], "y_pred": [np.min, np.mean, np.max]}).reset_index()
    bins_table.columns = [
        "Bin number",
        "Amount of objects",
        "Mean target",
        "Min probability",
        "Average probability",
        "Max probability",
    ]
    return bins_table.to_html(index=False)


# Regression plots:
def plot_target_distribution(data, path):

    sns.set(style="whitegrid", font_scale=1.5)
    g = sns.displot(
        data,
        bins=100,
        weights="count",
        x="Target value",
        row="source",
        height=9,
        aspect=1.5,
        kde=True,
        color="m",
        facet_kws=dict(margin_titles=True),
    )
    g.fig.suptitle("Target distribution")
    g.fig.tight_layout()
    g.fig.subplots_adjust(top=0.95)

    g.fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_error_hist(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(16, 10))

    sns.kdeplot(data["err"], shade=True, color="m", ax=ax, weights=data["count"])
    ax.set_xlabel("Error = y_pred - y_true")
    ax.set_ylabel("Density")
    ax.set_title("Error histogram")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


def plot_reg_scatter(data, path):
    sns.set(style="whitegrid", font_scale=1.5)
    g = sns.jointplot(x="y_pred", y="y_true", data=data, kind="reg", truncate=False, color="m", height=14)
    g.fig.suptitle("Scatter plot by subsample (2000 points)")
    g.fig.tight_layout()
    g.fig.subplots_adjust(top=0.95)

    g.fig.savefig(path, bbox_inches="tight")
    plt.close()


# Multiclass plots:


def plot_confusion_matrix(data: DenseMatrix, path):
    arr = data.toArray()
    cmat: "np.array" = arr / arr.sum(axis=1, keepdims=True)
    # cmat: "np.array" = confusion_matrix(data["y_true"], data["y_pred"], normalize="true")
    sns.set(style="whitegrid", font_scale=1.5)
    fig, ax = plt.subplots(figsize=(16, 12))

    sns.heatmap(cmat, annot=True, linewidths=0.5, cmap="Purples", ax=ax)
    ax.set_xlabel("y_pred")
    ax.set_ylabel("y_true")
    ax.set_title("Confusion matrix")

    fig.savefig(path, bbox_inches="tight")
    plt.close()


# Feature importance
def plot_feature_importance(feat_imp, path, features_max=100):
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(figsize=(16, features_max / 2.5))
    sns.barplot(x="Importance", y="Feature", data=feat_imp[:features_max], ax=axs, color="m")
    plt.savefig(path, bbox_inches="tight")
    plt.close()


class SparkReportDeco:
    """
    Decorator to wrap :class:`~lightautoml.automl.base.AutoML` class to generate html report on ``fit_predict`` and ``predict``.

    Example:

        >>> report_automl = SparkReportDeco(output_path="output_path", report_file_name="report_file_name")(automl).
        >>> report_automl.fit_predict(train_data)
        >>> report_automl.predict(test_data)

    Report will be generated at output_path/report_file_name automatically.

    Warning:
         Do not use it just to inference (if you don"t need report), because:

            - It needs target variable to calc performance metrics.
            - It takes additional time to generate report.
            - Dump of decorated automl takes more memory to store.

    To get unwrapped fitted instance to pickle
    and inferecne access ``report_automl.model`` attribute.

    """

    @property
    def model(self):
        """Get unwrapped model.

        Returns:
            model.

        """
        return self._model

    @property
    def mapping(self):
        return self._model.reader.class_mapping

    @property
    def task(self):
        return self._model.reader.task._name

    def __init__(self, *args, **kwargs):
        """

        Note:
            Valid kwargs are:

                - output_path: Folder with report files.
                - report_file_name: Name of main report file.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        """
        if not kwargs:
            kwargs = {}

        # default params
        self.fi_params = {"method": "fast", "n_sample": 100_000}
        self.interpretation_params = {
            "top_n_features": 5,
            "top_n_categories": 10,
            "top_n_classes": 10,
            "n_bins": 5,
            "datetime_level": "year",
            "n_sample": 100_000,
            "ice_fraction": 1.0,
            "ice_fraction_seed": 42,
        }

        fi_input_params = kwargs.get("fi_params", {})
        self.fi_params.update(fi_input_params)
        interpretation_input_params = kwargs.get("interpretation_params", {})
        self.interpretation_params.update(interpretation_input_params)
        self.interpretation = kwargs.get("interpretation", False)

        self.n_bins = kwargs.get("n_bins", 20)
        self.template_path = kwargs.get("template_path", os.path.join(base_dir, "spark_report_templates/"))
        self.output_path = kwargs.get("output_path", "sparklama_report/")
        self.report_file_name = kwargs.get("report_file_name", "sparklama_interactive_report.html")
        self.pdf_file_name = kwargs.get("pdf_file_name", None)

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path, exist_ok=True)

        self._base_template_path = "lama_base_template.html"
        self._model_section_path = "model_section.html"
        self._train_set_section_path = "train_set_section.html"
        self._results_section_path = "results_section.html"
        self._fi_section_path = "feature_importance_section.html"
        self._interpretation_section_path = "interpretation_section.html"
        self._interpretation_subsection_path = "interpretation_subsection.html"

        self._inference_section_path = {
            "binary": "binary_inference_section.html",
            "reg": "reg_inference_section.html",
            "multiclass": "multiclass_inference_section.html",
        }

        self.title = "Spark-LAMA report"
        if self.interpretation:
            self.sections_order = [
                "intro",
                "model",
                "train_set",
                "fi",
                "interpretation",
                "results",
            ]
            self._interpretation_top = []
        else:
            self.sections_order = ["intro", "model", "train_set", "fi", "results"]
        self._sections = dict()
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []

        self.generate_report()

    def __call__(self, model):
        self._model = model

        # add informataion to report
        self._model_name = model.__class__.__name__
        self._model_parameters = json2html.convert(extract_params(model))
        self._model_summary = None

        self._sections = dict()
        self._sections["intro"] = "<p>This report was generated automatically.</p>"
        self._model_results = []
        self._n_test_sample = 0

        self._generate_model_section()
        self.generate_report()
        return self

    def _binary_classification_details(
        self, data, positive_rate, true_labels_col_name, scores_col_name, predicted_labels_col_name
    ):
        prec, rec, F1 = plot_pie_f1_metric(
            data.select(
                F.col(predicted_labels_col_name).astype("double"), F.col(true_labels_col_name).astype("double")
            ).rdd,
            path=os.path.join(self.output_path, self._inference_content["pie_f1_metric"]),
        )

        # Done
        auc_score, ap_score, rounded_data = plot_curves(
            input_data=data,
            scores_col_name=scores_col_name,
            positive_rate=positive_rate,
            true_labels_col_name=true_labels_col_name,
            roc_path=os.path.join(self.output_path, self._inference_content["roc_curve"]),
            pr_path=os.path.join(self.output_path, self._inference_content["pr_curve"]),
        )

        # Done
        plot_distribution_of_logits(
            input_data=data,
            path=os.path.join(self.output_path, self._inference_content["distribution_of_logits"]),
            scores_col_name=scores_col_name,
            true_labels_col_name=true_labels_col_name,
        )

        return auc_score, prec, rec, F1

    def _regression_details(self, data, true_values_col_name, predictions_col_name):

        true_col = F.col(true_values_col_name)
        pred_col = F.col(predictions_col_name)

        true_max, pred_max, variance_y_true = data.select(
            F.max(F.abs(true_col)), F.max(F.abs(pred_col)), F.var_pop(true_col)
        ).first()

        _data = data.select(
            (F.round(true_col / float(true_max), 3) * true_max).alias(true_values_col_name),
            (F.round(pred_col / float(pred_max), 3) * pred_max).alias(predictions_col_name),
        ).cache()

        _pred = _data.groupBy(pred_col).count().toPandas()
        _true = _data.groupBy(true_col).count().toPandas()

        pred_data = pd.DataFrame({"Target value": _pred[predictions_col_name], "count": _pred["count"]})
        pred_data["source"] = "y_pred"

        true_data = pd.DataFrame({"Target value": _true[true_values_col_name], "count": _true["count"]})
        true_data["source"] = "y_true"

        target_distribution_data = pd.concat([pred_data, true_data], ignore_index=True)

        plot_target_distribution(
            target_distribution_data,
            path=os.path.join(self.output_path, self._inference_content["target_distribution"]),
        )

        errs = data.select((pred_col - true_col).alias("err"))

        err_max, err_min = errs.select(F.max(F.col("err")), F.min(F.col("err"))).first()

        val = err_max if abs(err_max) > abs(err_min) else err_min

        err_data = (
            errs.select((F.round(F.col("err") / val, 3) * val).alias("err")).groupBy(F.col("err")).count().toPandas()
        )

        plot_error_hist(
            err_data,
            path=os.path.join(self.output_path, self._inference_content["error_hist"]),
        )

        plot_reg_scatter(
            pd.DataFrame(
                data.select(true_col.alias("y_true"), pred_col.alias("y_pred")).rdd.takeSample(
                    withReplacement=False, num=2000, seed=42
                ),
                columns=["y_true", "y_pred"],
            ),
            path=os.path.join(self.output_path, self._inference_content["scatter_plot"]),
        )

        # metrics
        metrics = RegressionMetrics(
            predictionAndObservations=data.select(
                F.col(predictions_col_name).astype("double"), F.col(true_values_col_name).astype("double")
            ).rdd
        )

        mean_ae = metrics.meanAbsoluteError

        median_ae = data.select(
            F.percentile_approx(
                F.abs(F.col(predictions_col_name).astype("double") - F.col(true_values_col_name).astype("double")), 0.5
            )
        ).first()[0]

        mse = metrics.meanSquaredError

        r2 = metrics.r2

        # Due to Spark returns not a fraction but an absolute value, we have to calculate
        # explained variance score by ourselves.
        # evs = metrics.explainedVariance

        # Fraction of variance explained (FVE) = 1 - Fraction of variance unexplained (FVU).
        # FVU = MSE / Variance[Y] [https://en.wikipedia.org/wiki/Fraction_of_variance_unexplained]
        fve = 1 - (mse / variance_y_true)

        return mean_ae, median_ae, mse, r2, fve

    def _multiclass_details(self, data, predicted_labels_col_name, true_labels_col_name):

        true_labels_col = F.col(true_labels_col_name)

        metrics = MulticlassMetrics(
            predictionAndLabels=data.select(
                F.col(predicted_labels_col_name).astype("double"), true_labels_col.astype("double")
            ).rdd
        )

        # tn, fp, fn, tp = metrics.confusionMatrix().values
        labels_counts: pd.DataFrame = data.select(true_labels_col).groupby(true_labels_col).count().toPandas()
        labels_counts.sort_values(by=[true_labels_col_name], ascending=True, inplace=True)

        total_labels_number = len(labels_counts["count"])

        # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html
        p_micro = metrics.accuracy
        p_macro_sum = 0.0
        for label in labels_counts[true_labels_col_name]:
            p_macro_sum += metrics.precision(float(label))
        p_macro = p_macro_sum / total_labels_number
        p_weighted = metrics.weightedPrecision

        r_micro = metrics.accuracy
        r_macro_sum = 0.0
        for label in labels_counts[true_labels_col_name]:
            r_macro_sum += metrics.recall(float(label))
        r_macro = r_macro_sum / total_labels_number
        r_weighted = metrics.weightedRecall

        f_micro = metrics.accuracy
        f_macro_sum = 0.0
        for label in labels_counts[true_labels_col_name]:
            f_macro_sum += metrics.fMeasure(float(label))
        f_macro = f_macro_sum / total_labels_number
        f_weighted = metrics.weightedFMeasure()

        # classification report for features
        if self.mapping:
            classes = sorted(self.mapping, key=self.mapping.get)
        else:
            classes = np.arange(self._N_classes)

        p = [metrics.precision(float(label)) for label in labels_counts[true_labels_col_name]]
        r = [metrics.recall(float(label)) for label in labels_counts[true_labels_col_name]]
        f = [metrics.fMeasure(float(label)) for label in labels_counts[true_labels_col_name]]
        s = list(labels_counts["count"])

        # p, r, f, s = precision_recall_fscore_support(y_true, y_pred)
        cls_report = pd.DataFrame(
            {
                "Class name": classes,
                "Precision": p,
                "Recall": r,
                "F1-score": f,
                "Support": s,
            }
        )
        self._inference_content["classification_report"] = cls_report.to_html(
            index=False, float_format="{:.4f}".format, justify="left"
        )

        plot_confusion_matrix(
            metrics.confusionMatrix(),
            path=os.path.join(self.output_path, self._inference_content["confusion_matrix"]),
        )

        return [
            p_micro,
            p_macro,
            p_weighted,
            r_micro,
            r_macro,
            r_weighted,
            f_micro,
            f_macro,
            f_weighted,
        ]

    def _collect_data(
        self, preds: SparkDataset, sample, true_values_col_name, raw_predictions_col_name, predictions_col_name
    ) -> SparkDataFrame:
        # TODO: SPARK-LAMA hack, replace it later.
        raw_pred_col = next(c for c in preds.data.columns if c.startswith("prediction"))

        @pandas_udf("double")
        def argmax(s: pd.Series) -> pd.Series:
            return s.map(np.argmax)

        if preds.task.name == "multiclass":
            raw_prediction_col = vector_to_array(raw_pred_col).alias(raw_predictions_col_name)
            prediction_col = argmax(vector_to_array(raw_pred_col)).alias(predictions_col_name)
        elif preds.task.name == "binary":
            raw_prediction_col = vector_to_array(raw_pred_col).getItem(1).alias(raw_predictions_col_name)
            prediction_col = argmax(vector_to_array(raw_pred_col)).alias(predictions_col_name)
        else:
            raw_prediction_col = F.col(raw_pred_col).alias(raw_predictions_col_name)
            prediction_col = F.col(raw_pred_col).alias(predictions_col_name)

        df = preds.data.where(~F.isnull(raw_pred_col)).select(
            F.col(SparkDataset.ID_COLUMN),
            F.col(preds.target_column).alias(true_values_col_name),
            raw_prediction_col,
            prediction_col,
        )

        return df

    def fit_predict(self, *args, **kwargs):
        """Wrapped ``automl.fit_predict`` method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: Arguments.
            **kwargs: Additional parameters.

        Returns:
            OOF predictions.

        """
        # TODO: parameters parsing in general case

        preds: SparkDataset = self._model.fit_predict(*args, **kwargs)

        true_values_col_name = "y_true"
        scores_col_name = "raw"
        predictions_col_name = "label"

        train_data: DataFrame = kwargs["train_data"] if "train_data" in kwargs else args[0]
        valid_data: Optional[DataFrame] = kwargs.get("valid_data", None)
        input_roles = kwargs["roles"] if "roles" in kwargs else args[1]
        self._target = input_roles["target"]

        # data = self._get_mock_data()
        if valid_data is None:
            data = self._collect_data(preds, train_data, true_values_col_name, scores_col_name, predictions_col_name)
        else:
            data = self._collect_data(preds, valid_data, true_values_col_name, scores_col_name, predictions_col_name)

        self._inference_content = {}
        if self.task == "binary":
            # filling for html
            self._inference_content = dict()
            self._inference_content["roc_curve"] = "valid_roc_curve.png"
            self._inference_content["pr_curve"] = "valid_pr_curve.png"
            self._inference_content["pie_f1_metric"] = "valid_pie_f1_metric.png"
            self._inference_content["distribution_of_logits"] = "valid_distribution_of_logits.png"
            # graphics and metrics
            _, self._F1_thresh, positive_rate = f1_score_w_co(
                data, true_labels_col_name=true_values_col_name, scores_col_name=scores_col_name
            )
            auc_score, prec, rec, F1 = self._binary_classification_details(
                data,
                positive_rate,
                true_labels_col_name=true_values_col_name,
                scores_col_name=scores_col_name,
                predicted_labels_col_name=predictions_col_name,
            )
            # update model section
            evaluation_parameters = ["AUC-score", "Precision", "Recall", "F1-score"]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [auc_score, prec, rec, F1],
                }
            )
        elif self.task == "reg":
            # filling for html
            self._inference_content["target_distribution"] = "valid_target_distribution.png"
            self._inference_content["error_hist"] = "valid_error_hist.png"
            self._inference_content["scatter_plot"] = "valid_scatter_plot.png"
            # graphics and metrics

            mean_ae, median_ae, mse, r2, evs = self._regression_details(
                data, true_values_col_name, predictions_col_name
            )
            # model section
            evaluation_parameters = [
                "Mean absolute error",
                "Median absolute error",
                "Mean squared error",
                "R^2 (coefficient of determination)",
                "Explained variance",
            ]
            self._model_summary = pd.DataFrame(
                {
                    "Evaluation parameter": evaluation_parameters,
                    "Validation sample": [mean_ae, median_ae, mse, r2, evs],
                }
            )
        elif self.task == "multiclass":
            self._N_classes = train_data.select(F.count_distinct(F.col(self._target))).first()[0]

            self._inference_content["confusion_matrix"] = "valid_confusion_matrix.png"

            index_names = np.array([["Precision", "Recall", "F1-score"], ["micro", "macro", "weighted"]])
            index = pd.MultiIndex.from_product(index_names, names=["Evaluation metric", "Average"])

            summary = self._multiclass_details(data, predictions_col_name, true_values_col_name)
            self._model_summary = pd.DataFrame({"Validation sample": summary}, index=index)

        self._inference_content["title"] = "Results on validation sample"

        self._generate_model_section()

        # generate train data section
        self._train_data_overview = self._data_genenal_info(train_data)
        self._describe_roles(train_data)
        self._describe_dropped_features(train_data)
        self._generate_train_set_section()

        # generate fit_predict section
        self._generate_inference_section()

        # generate feature importance and interpretation sections
        self._generate_fi_section(valid_data)

        if self.interpretation:
            self._generate_interpretation_section(valid_data)

        self.generate_report()
        return preds

    def predict(self, *args, **kwargs):
        """Wrapped automl.predict method.

        Valid args, kwargs are the same as wrapped automl.

        Args:
            *args: arguments.
            **kwargs: additional parameters.

        Returns:
            predictions.

        """
        self._n_test_sample += 1
        # get predictions
        test_preds = self._model.predict(*args, **kwargs)

        true_values_col_name = "y_true"
        scores_col_name = "raw"
        predictions_col_name = "label"

        test_data = kwargs["test"] if "test" in kwargs else args[0]

        data = self._collect_data(
            test_preds,
            test_data,
            true_values_col_name=true_values_col_name,
            raw_predictions_col_name=scores_col_name,
            predictions_col_name=predictions_col_name,
        )

        if self.task == "binary":
            # filling for html
            self._inference_content = {}
            self._inference_content["roc_curve"] = "test_roc_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pr_curve"] = "test_pr_curve_{}.png".format(self._n_test_sample)
            self._inference_content["pie_f1_metric"] = "test_pie_f1_metric_{}.png".format(self._n_test_sample)
            self._inference_content["bins_preds"] = "test_bins_preds_{}.png".format(self._n_test_sample)
            self._inference_content["preds_distribution_by_bins"] = "test_preds_distribution_by_bins_{}.png".format(
                self._n_test_sample
            )
            self._inference_content["distribution_of_logits"] = "test_distribution_of_logits_{}.png".format(
                self._n_test_sample
            )
            # graphics and metrics
            true_values_col = F.col(true_values_col_name)

            positive_rate = data.select(F.sum(true_values_col) / F.count(true_values_col)).first()[0]
            auc_score, prec, rec, F1 = self._binary_classification_details(
                data,
                positive_rate=positive_rate,
                true_labels_col_name=true_values_col_name,
                scores_col_name=scores_col_name,
                predicted_labels_col_name=predictions_col_name,
            )

            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = [
                    auc_score,
                    prec,
                    rec,
                    F1,
                ]
            else:
                self._model_summary["Test sample"] = [auc_score, prec, rec, F1]

        elif self.task == "reg":
            # filling for html

            predictions_col_name = scores_col_name

            self._inference_content = {}
            self._inference_content["target_distribution"] = "test_target_distribution_{}.png".format(
                self._n_test_sample
            )
            self._inference_content["error_hist"] = "test_error_hist_{}.png".format(self._n_test_sample)
            self._inference_content["scatter_plot"] = "test_scatter_plot_{}.png".format(self._n_test_sample)
            # graphics
            mean_ae, median_ae, mse, r2, evs = self._regression_details(
                data, true_values_col_name, predictions_col_name
            )
            # update model section
            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = [
                    mean_ae,
                    median_ae,
                    mse,
                    r2,
                    evs,
                ]
            else:
                self._model_summary["Test sample"] = [mean_ae, median_ae, mse, r2, evs]

        elif self.task == "multiclass":
            self._inference_content["confusion_matrix"] = "test_confusion_matrix_{}.png".format(self._n_test_sample)
            test_summary = self._multiclass_details(
                data, predicted_labels_col_name=predictions_col_name, true_labels_col_name=true_values_col_name
            )
            if self._n_test_sample >= 2:
                self._model_summary["Test sample {}".format(self._n_test_sample)] = test_summary
            else:
                self._model_summary["Test sample"] = test_summary

        # layout depends on number of test samples
        if self._n_test_sample >= 2:
            self._inference_content["title"] = "Results on test sample {}".format(self._n_test_sample)

        else:
            self._inference_content["title"] = "Results on test sample"

        # update model section
        self._generate_model_section()
        # generate predict section
        self._generate_inference_section()

        self.generate_report()
        return test_preds

    def _generate_fi_section(self, valid_data: Optional[DataFrame]):
        if self.fi_params["method"] == "accurate" and valid_data is None:
            self.fi_params["method"] = "fast"
            warnings.warn(
                "You must set valid_data with 'accurate' feature importance method. Changed to 'fast' automatically."
            )

        self.feat_imp = self._model.get_feature_scores(
            calc_method=self.fi_params["method"], data=valid_data, silent=False
        )
        if self.feat_imp is None:
            fi_path = None
        else:
            fi_path = "feature_importance.png"
            plot_feature_importance(self.feat_imp, path=os.path.join(self.output_path, fi_path))
        # add to _sections
        fi_content = {
            "fi_method": self.fi_params["method"],
            "feature_importance": fi_path,
        }
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        fi_section = env.get_template(self._fi_section_path).render(fi_content)
        self._sections["fi"] = fi_section

    def _generate_interpretation_content(self, test_data):
        self._interpretation_content = {}
        if test_data is None:
            self._interpretation_content["interpretation_top"] = None
            return
        if self.feat_imp is None:
            interpretation_feat_list = list(self._model.reader._roles.keys())[
                : self.interpretation_params["top_n_features"]
            ]
        else:
            interpretation_feat_list = self.feat_imp["Feature"].values[: self.interpretation_params["top_n_features"]]
        for feature_name in interpretation_feat_list:
            interpretaton_subsection = dict()
            interpretaton_subsection["feature_name"] = feature_name
            interpretaton_subsection["feature_interpretation_plot"] = feature_name + "_interpretation.png"
            self._plot_pdp(
                test_data,
                feature_name,
                path=os.path.join(
                    self.output_path,
                    interpretaton_subsection["feature_interpretation_plot"],
                ),
            )
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            interpretation_subsection = env.get_template(self._interpretation_subsection_path).render(
                interpretaton_subsection
            )
            self._interpretation_top.append(interpretation_subsection)
            print(f"Interpretation info for {feature_name} appended")
        self._interpretation_content["interpretation_top"] = self._interpretation_top

    def _generate_interpretation_section(self, test_data: DataFrame):
        self._generate_interpretation_content(test_data)
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        interpretation_section = env.get_template(self._interpretation_section_path).render(
            self._interpretation_content
        )
        self._sections["interpretation"] = interpretation_section

    def _plot_pdp(self, test_data: DataFrame, feature_name, path):
        feature_role = self._model.reader._roles[feature_name].name
        # I. Count interpretation
        print("Calculating interpretation for {}:".format(feature_name))
        grid, ys, counts = self._model.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=self.interpretation_params["n_bins"],
            top_n_categories=self.interpretation_params["top_n_categories"],
            datetime_level=self.interpretation_params["datetime_level"],
            ice_fraction=self.interpretation_params["ice_fraction"],
            ice_fraction_seed=self.interpretation_params["ice_fraction_seed"],
        )

        HISTOGRAM_DATA_ROWS_LIMIT = 2000
        rows_count = test_data.count()
        if rows_count > HISTOGRAM_DATA_ROWS_LIMIT:
            fraction = HISTOGRAM_DATA_ROWS_LIMIT / rows_count
            test_data = test_data.sample(fraction=fraction)

        if self._model.reader._roles[feature_name].name == "Numeric":
            test_data: pd.DataFrame = test_data.select(F.col(feature_name).cast("double")).toPandas()
        else:
            test_data: pd.DataFrame = test_data.select(feature_name).toPandas()

        # II. Plot pdp
        sns.set(style="whitegrid", font_scale=1.5)
        fig, axs = plt.subplots(2, 1, figsize=(16, 12), gridspec_kw={"height_ratios": [3, 1]})
        axs[0].set_title("PDP plot: " + feature_name)
        n_classes = ys[0].shape[1]
        if n_classes == 1:
            data = pd.concat(
                [pd.DataFrame({"x": grid[i], "y": ys[i].ravel()}) for i, _ in enumerate(grid)]
            ).reset_index(drop=True)
            if feature_role in ["Numeric", "Datetime"]:
                g0 = sns.lineplot(data=data, x="x", y="y", ax=axs[0], color="m")
            else:
                g0 = sns.boxplot(data=data, x="x", y="y", ax=axs[0], showfliers=False, color="m")
        else:
            if self.mapping:
                classes = sorted(self.mapping, key=self.mapping.get)[: self.interpretation_params["top_n_classes"]]
            else:
                classes = np.arange(min(n_classes, self.interpretation_params["top_n_classes"]))
            data = pd.concat(
                [
                    pd.DataFrame({"x": grid[i], "y": ys[i][:, k], "class": name})
                    for i, _ in enumerate(grid)
                    for k, name in enumerate(classes)
                ]
            ).reset_index(drop=True)
            if self._model.reader._roles[feature_name].name in ["Numeric", "Datetime"]:
                g0 = sns.lineplot(data=data, x="x", y="y", hue="class", ax=axs[0])
            else:
                g0 = sns.boxplot(data=data, x="x", y="y", hue="class", ax=axs[0], showfliers=False)
        g0.set(ylabel="y_pred")
        # III. Plot distribution
        counts = np.array(counts) / sum(counts)
        if feature_role == "Numeric":
            g0.set(xlabel="feature value")
            g1 = sns.histplot(test_data[feature_name], kde=True, color="gray", ax=axs[1])
        elif feature_role == "Category":
            g0.set(xlabel=None)
            axs[0].set_xticklabels(grid, rotation=90)
            g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
        else:
            g0.set(xlabel=self.interpretation_params["datetime_level"])
            g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
        g1.set(xlabel=None)
        g1.set(ylabel="Frequency")
        g1.set(xticklabels=[])
        # IV. Save picture
        plt.tight_layout()
        fig.savefig(path, bbox_inches="tight")
        plt.close()

    def _data_genenal_info(self, data):
        general_info = pd.DataFrame(columns=["Parameter", "Value"])
        general_info.loc[0] = ("Number of records", data.count())
        general_info.loc[1] = ("Total number of features", len(data.columns))
        general_info.loc[2] = ("Used features", len(self._model.reader._used_features))
        general_info.loc[3] = (
            "Dropped features",
            len(self._model.reader._dropped_features),
        )
        # general_info.loc[4] = ("Number of positive cases", np.sum(data[self._target] == 1))
        # general_info.loc[5] = ("Number of negative cases", np.sum(data[self._target] == 0))
        return general_info.to_html(index=False, justify="left")

    def _get_column_nan_ratio(self, column: Column, column_name: str, total_count: int) -> Column:
        return (
            F.sum(F.when(F.isnan(column), 1).otherwise(F.when(F.isnull(column), 1).otherwise(0))) / total_count
        ).alias(
            f"nanratio_{column_name}"
        )  # column._jc.toString()

    def _exclude_nan_values(self, column: Column) -> Column:
        return F.when(F.isnan(column), None).otherwise(column)

    def _get_column_min(self, column: Column, column_name: str, astype: Optional[str] = None) -> Column:
        if astype:
            column = column.astype(astype)
        return F.min(column).alias(f"min_{column_name}")

    def _get_column_percentile(
        self, column: Column, column_name: str, percentile: float, astype: Optional[str] = None
    ) -> Column:
        if astype:
            column = column.astype(astype)
        return F.percentile_approx(column, percentile).alias(f"perc{percentile}_{column_name}")

    def _get_column_avg(self, column: Column, column_name: str, astype: Optional[str] = None) -> Column:
        if astype:
            column = column.astype(astype)
        return F.avg(column).alias(f"avg_{column_name}")

    def _get_column_max(self, column: Column, column_name: str, astype: Optional[str] = None) -> Column:
        if astype:
            column = column.astype(astype)
        return F.max(column).alias(f"max_{column_name}")

    def _describe_roles(self, train_data):

        # detect feature roles
        roles = self._model.reader._roles
        numerical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Numeric"]
        categorical_features = [feat_name for feat_name in roles if roles[feat_name].name == "Category"]
        datetime_features = [feat_name for feat_name in roles if roles[feat_name].name == "Datetime"]

        total_count = train_data.count()
        # numerical roles
        numerical_features_df = []

        columns_to_select = list()

        for feature_name in numerical_features:
            column = F.col(feature_name)
            columns_to_select.append(self._get_column_nan_ratio(column, feature_name, total_count))

            filtered_column = self._exclude_nan_values(column)
            columns_to_select.append(self._get_column_min(filtered_column, feature_name, astype="double"))
            columns_to_select.append(self._get_column_percentile(filtered_column, feature_name, 0.25, astype="double"))
            columns_to_select.append(self._get_column_avg(filtered_column, feature_name, astype="double"))
            columns_to_select.append(self._get_column_percentile(filtered_column, feature_name, 0.5, astype="double"))
            columns_to_select.append(self._get_column_percentile(filtered_column, feature_name, 0.75, astype="double"))
            columns_to_select.append(self._get_column_max(filtered_column, feature_name, astype="double"))

        for feature_name in categorical_features:
            columns_to_select.append(self._get_column_nan_ratio(F.col(feature_name), feature_name, total_count))

        for feature_name in datetime_features:
            column = F.col(feature_name)
            columns_to_select.append(self._get_column_nan_ratio(column, feature_name, total_count))

            filtered_column = self._exclude_nan_values(column)
            columns_to_select.append(self._get_column_min(filtered_column, feature_name))
            columns_to_select.append(self._get_column_max(filtered_column, feature_name))

        stat_data: pd.DataFrame = train_data.select(*columns_to_select).toPandas()
        # +-----------------+------------+-----+------------+------------+
        # | nanratio_<col1> | min_<col1> | ... | avg_<colN> | max_<colN> |
        # +-----------------+------------+-----+------------+------------+
        # |             0.0 |      162.0 | ... |    113.274 |      999.0 |
        # +-----------------+------------+-----+------------+------------+

        for feature_name in numerical_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(stat_data[f"nanratio_{feature_name}"][0])
            item["min"] = stat_data[f"min_{feature_name}"][0]
            item["quantile_25"] = stat_data[f"perc0.25_{feature_name}"][0]
            item["average"] = stat_data[f"avg_{feature_name}"][0]
            item["median"] = stat_data[f"perc0.5_{feature_name}"][0]
            item["quantile_75"] = stat_data[f"perc0.75_{feature_name}"][0]
            item["max"] = stat_data[f"max_{feature_name}"][0]
            numerical_features_df.append(item)
        if len(numerical_features_df) == 0:
            self._numerical_features_table = None
        else:
            self._numerical_features_table = pd.DataFrame(numerical_features_df).to_html(
                index=False, float_format="{:.2f}".format, justify="left"
            )

        indexer = LAMLStringIndexer(
            inputCols=categorical_features,
            outputCols=[f"{col}_out" for col in categorical_features],
            minFreqs=[0 for _ in categorical_features],
            handleInvalid="skip",
            defaultValue=0.0,
            freqLabel=True,
        )

        indexer_model: LAMLStringIndexerModel = indexer.fit(train_data)

        encodings = indexer_model.labelsArray

        # categorical roles
        categorical_features_df = []
        for feature_name, enc in zip(categorical_features, encodings):
            sorted_enc: list = [
                pair
                for pair in sorted(enc, key=lambda x: int(x[1]), reverse=True)
                if pair[0] is not None and pair[0] != "None"
            ]

            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(stat_data[f"nanratio_{feature_name}"][0])
            item["Number of unique values"] = len(sorted_enc)
            item["Most frequent value"] = sorted_enc[0][0]
            item["Occurance of most frequent"] = "{:.1f}%".format(100 * (int(sorted_enc[0][1]) / float(total_count)))
            item["Least frequent value"] = sorted_enc[-1][0]
            item["Occurance of least frequent"] = "{:.1f}%".format(100 * (int(sorted_enc[-1][1]) / float(total_count)))
            categorical_features_df.append(item)
        if len(categorical_features_df) == 0:
            self._categorical_features_table = None
        else:
            self._categorical_features_table = pd.DataFrame(categorical_features_df).to_html(
                index=False, justify="left"
            )

        # datetime roles
        datetime_features_df = []
        for feature_name in datetime_features:
            item = {"Feature name": feature_name}
            item["NaN ratio"] = "{:.4f}".format(stat_data[f"nanratio_{feature_name}"][0])
            item["min"] = stat_data[f"min_{feature_name}"][0]
            item["min"] = stat_data[f"max_{feature_name}"][0]
            item["base_date"] = self._model.reader._roles[feature_name].base_date
            datetime_features_df.append(item)
        if len(datetime_features_df) == 0:
            self._datetime_features_table = None
        else:
            self._datetime_features_table = pd.DataFrame(datetime_features_df).to_html(index=False, justify="left")

    def _describe_dropped_features(self, train_data):
        total_count = train_data.count()
        self._max_nan_rate = self._model.reader.max_nan_rate
        self._max_constant_rate = self._model.reader.max_constant_rate
        self._features_dropped_list = self._model.reader._dropped_features
        # dropped features table
        dropped_list = [col for col in self._features_dropped_list if col != self._target]
        if len(dropped_list) == 0:
            self._dropped_features_table = None
        else:
            dropped_nan_ratio = train_data.select(
                *[
                    (
                        F.sum(
                            F.when(F.isnan(F.col(col_name)), 1).otherwise(
                                F.when(F.isnull(F.col(col_name)), 1).otherwise(0)
                            )
                        )
                        / total_count
                    ).alias(col_name)
                    for col_name in dropped_list
                ]
            ).toPandas()

            indexer = LAMLStringIndexer(
                inputCols=dropped_list,
                outputCols=[f"{col}_out" for col in dropped_list],
                minFreqs=[0 for _ in dropped_list],
                handleInvalid="skip",
                defaultValue=0.0,
                freqLabel=True,
            )

            indexer_model: LAMLStringIndexerModel = indexer.fit(train_data)

            encodings = indexer_model.labelsArray  # list of string tuples ('key', 'count'), example: ('key', '11')
            dropped_most_occured = [
                int(sorted(enc, key=lambda x: int(x[1]), reverse=True)[0][1]) / float(total_count) for enc in encodings
            ]

            dropped_features_table = pd.DataFrame(
                [
                    {"Название переменной": var_name, "nan_rate": nan_ratio, "constant_rate": most_occured}
                    for var_name, nan_ratio, most_occured in zip(
                        dropped_nan_ratio, dropped_nan_ratio.iloc[0], dropped_most_occured
                    )
                ]
            )
            self._dropped_features_table = dropped_features_table.to_html(index=False, justify="left")

    def _generate_model_section(self):
        model_summary = None
        if self._model_summary is not None:
            model_summary = self._model_summary.to_html(
                index=self.task == "multiclass",
                justify="left",
                float_format="{:.4f}".format,
            )

        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        model_section = env.get_template(self._model_section_path).render(
            model_name=self._model_name,
            model_parameters=self._model_parameters,
            model_summary=model_summary,
        )
        self._sections["model"] = model_section

    def _generate_train_set_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        train_set_section = env.get_template(self._train_set_section_path).render(
            train_data_overview=self._train_data_overview,
            numerical_features_table=self._numerical_features_table,
            categorical_features_table=self._categorical_features_table,
            datetime_features_table=self._datetime_features_table,
            target=self._target,
            max_nan_rate=self._max_nan_rate,
            max_constant_rate=self._max_constant_rate,
            dropped_features_table=self._dropped_features_table,
        )
        self._sections["train_set"] = train_set_section

    def _generate_inference_section(self):
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        inference_section = env.get_template(self._inference_section_path[self.task]).render(self._inference_content)
        self._model_results.append(inference_section)

    def _generate_results_section(self):
        if self._model_results:
            env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
            results_section = env.get_template(self._results_section_path).render(model_results=self._model_results)
            self._sections["results"] = results_section

    def generate_report(self):
        # collection sections
        self._generate_results_section()
        sections_list = []
        for sec_name in self.sections_order:
            if sec_name in self._sections:
                sections_list.append(self._sections[sec_name])
        # put sections inside
        env = Environment(loader=FileSystemLoader(searchpath=self.template_path))
        report = env.get_template(self._base_template_path).render(
            title=self.title, sections=sections_list, pdf=self.pdf_file_name
        )

        with open(os.path.join(self.output_path, self.report_file_name), "w", encoding="utf-8") as f:
            f.write(report)

        if self.pdf_file_name:
            try:
                from weasyprint import HTML

                HTML(string=report, base_url=self.output_path).write_pdf(
                    os.path.join(self.output_path, self.pdf_file_name)
                )
            except ModuleNotFoundError:
                print("Can't generate PDF report: check manual for installing pdf extras.")



from copy import deepcopy
from typing import Dict, Iterator, Optional, Sequence, List
from collections import defaultdict, OrderedDict
from itertools import chain, combinations
from datetime import datetime
import holidays
import numpy as np
import pandas as pd
from pyspark.sql import functions as F, types as SparkTypes, DataFrame as SparkDataFrame
from pyspark.sql.pandas.functions import pandas_udf
from pyspark.sql.types import IntegerType

from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.roles import CategoryRole, NumericRole, ColumnRole
from sparklightautoml.mlwriters import CommonPickleMLReadable, CommonPickleMLWritable
from lightautoml.transformers.datetime import datetime_check, date_attrs

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.transformers.base import ObsoleteSparkTransformer, SparkBaseTransformer

from pyspark.ml import Transformer, Estimator
from pyspark.ml.param.shared import HasInputCols, HasOutputCols
from pyspark.ml.param.shared import TypeConverters, Param, Params


def get_unit_of_timestamp_column(seas: str, col: str):
    """Generates pyspark column to extract unit of time from timestamp

    Args:
        seas (str): unit of time: `y`(year), `m`(month), `d`(day),
        `wd`(weekday), `hour`(hour), `min`(minute), `sec`(second), `ms`(microsecond), `ns`(nanosecond)
        col (str): column name with datetime values
    """
    if seas == "y":
        return F.year(F.to_timestamp(F.col(col)))
    elif seas == "m":
        return F.month(F.to_timestamp(F.col(col)))
    elif seas == "d":
        return F.dayofmonth(F.to_timestamp(F.col(col)))
    # TODO SPARK-LAMA: F.dayofweek() starts numbering from another day.
    # Differs from pandas.Timestamp.weekday.
    # elif seas == 'wd':
    #     return F.dayofweek(F.to_timestamp(F.col(col)))
    elif seas == "hour":
        return F.hour(F.to_timestamp(F.col(col)))
    elif seas == "min":
        return F.minute(F.to_timestamp(F.col(col)))
    elif seas == "sec":
        return F.second(F.to_timestamp(F.col(col)))
    else:

        @pandas_udf(returnType=IntegerType())
        def get_timestamp_attr(arrs: Iterator[pd.Series]) -> Iterator[pd.Series]:
            for x in arrs:

                def convert_to_datetime(timestamp: int):
                    try:
                        date = pd.to_datetime(datetime.fromtimestamp(timestamp))
                    except:
                        date = datetime.now()
                    return date

                x = x.apply(lambda d: convert_to_datetime(d))
                yield getattr(x.dt, date_attrs[seas])

        return get_timestamp_attr(F.to_timestamp(F.col(col)).cast("long"))


class SparkDatetimeHelper:
    """
    Helper class for :class:`~sparklightautoml.transformers.datetime.SparkTimeToNumTransformer`,
    :class:`~sparklightautoml.transformers.datetime.SparkBaseDiffTransformer` and
    :class:`~sparklightautoml.transformers.datetime.SparkDateSeasonsTransformer`
    """

    basic_interval = "D"

    _interval_mapping = {
        "NS": 0.000000001,
        "MS": 0.001,
        "SEC": 1,
        "MIN": 60,
        "HOUR": 60 * 60,
        "D": 60 * 60 * 24,
        # FIXME SPARK-LAMA: Very rough rounding
        "M": 60 * 60 * 24 * 30,
        "Y": 60 * 60 * 24 * 365,
    }

    _fit_checks = (datetime_check,)
    _transform_checks = ()


class SparkTimeToNumTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Transforms datetime columns values to numeric values.
    """

    basic_time = "2020-01-01"
    _fname_prefix = "dtdiff"

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns: bool = False):
        output_cols = [f"{self._fname_prefix}__{f}" for f in input_cols]
        output_roles = {f: NumericRole(np.float32) for f in output_cols}
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        df = dataset

        new_cols = []
        for i, out_col in zip(self.getInputCols(), self.getOutputCols()):
            new_col = (
                (F.to_timestamp(F.col(i)).cast("long") - F.to_timestamp(F.lit(self.basic_time)).cast("long"))
                / self._interval_mapping[self.basic_interval]
            ).alias(out_col)
            new_cols.append(new_col)

        df = self._make_output_df(df, new_cols)

        return df


class SparkBaseDiffTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Basic conversion strategy, used in selection one-to-one transformers.
    Datetime converted to difference with basic_date.
    """

    _fname_prefix = "basediff"

    baseNames = Param(Params._dummy(), "baseNames", "base_names")

    diffNames = Param(Params._dummy(), "diffNames", "diff_names")

    basicInterval = Param(Params._dummy(), "basicInterval", "basic_interval")

    def __init__(
        self,
        input_roles: RolesDict,
        base_names: Sequence[str],
        diff_names: Sequence[str],
        basic_interval: Optional[str] = "D",
        do_replace_columns: bool = False,
    ):
        input_cols = list(base_names) + list(diff_names)

        self.base_names = base_names
        self.diff_names = diff_names
        self.basic_interval = basic_interval

        output_cols = [f"{self._fname_prefix}_{col}__{x}" for col in base_names for x in diff_names]

        output_roles = {col: NumericRole(dtype=np.float32) for col in output_cols}

        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)

        self.set(self.baseNames, self.base_names)
        self.set(self.diffNames, self.diff_names)
        self.set(self.basicInterval, self.basic_interval)

    def _transform(self, df: SparkDataFrame) -> SparkDataFrame:

        new_cols = [
            (
                (F.to_timestamp(F.col(dif)).cast("long") - F.to_timestamp(F.col(base)).cast("long"))
                / self._interval_mapping[self.basic_interval]
            ).alias(f"{self._fname_prefix}_{base}__{dif}")
            for base in self.base_names
            for dif in self.diff_names
        ]

        df = self._make_output_df(df, new_cols)

        return df


class SparkDateSeasonsTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Extracts unit of time from Datetime values and marks holiday dates.
    """

    _fname_prefix = "season"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: RolesDict,
        do_replace_columns: bool = False,
        output_role: Optional[ColumnRole] = None,
    ):
        self.output_role = output_role
        if output_role is None:
            self.output_role = CategoryRole(np.int32)

        self.transformations = OrderedDict()
        output_cols = []
        for col in input_cols:
            seas = input_roles[col].seasonality
            self.transformations[col] = seas
            for s in seas:
                output_cols.append(f"{self._fname_prefix}_{s}__{col}")
            if input_roles[col].country is not None:
                output_cols.append(f"{self._fname_prefix}_hol__{col}")

        output_roles = {f: deepcopy(self.output_role) for f in output_cols}

        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        df = dataset
        roles = self.getInputRoles()

        new_cols = []
        for col in self.getInputCols():
            fcol = F.to_timestamp(F.col(col)).cast("long")
            seas_cols = [
                (
                    F.when(F.isnan(fcol) | F.isnull(fcol), None)
                    .otherwise(get_unit_of_timestamp_column(seas, col))
                    .alias(f"{self._fname_prefix}_{seas}__{col}")
                )
                for seas in self.transformations[col]
            ]

            new_cols.extend(seas_cols)

            if roles[col].country is not None:

                @pandas_udf(returnType=IntegerType())
                def is_holiday(arrs: Iterator[pd.Series]) -> Iterator[pd.Series]:
                    for x in arrs:
                        x = x.apply(lambda d: datetime.fromtimestamp(d)).dt.normalize()
                        _holidays = holidays.CountryHoliday(
                            years=np.unique(x.dt.year.values),
                            country=roles[col].country,
                            prov=roles[col].prov,
                            state=roles[col].state,
                        )
                        yield x.isin(_holidays).astype(int)

                hol_col = is_holiday(fcol).alias(f"{self._fname_prefix}_hol__{col}")
                new_cols.append(hol_col)

        df = self._make_output_df(df, new_cols)

        return df



from typing import Union, List

from lightautoml.dataset.roles import NumericRole, Dtype
import numpy as np


class NumericVectorOrArrayRole(NumericRole):
    """
    Role that describe numeric vector or numeric array.
    """

    _name = "NumericOrArrayVector"

    def __init__(
        self,
        size: int,
        element_col_name_template: Union[str, List[str]],
        dtype: Dtype = np.float32,
        force_input: bool = False,
        prob: bool = False,
        discretization: bool = False,
        is_vector: bool = True,
    ):
        """
        Args:
            size: number of elements in every vector in this column
            element_col_name_template: string template to produce name for each element in the vector
            when array-to-columns transformation is neccessary
            dtype: type of the vector's elements
            force_input: Select a feature for training,
              regardless of the selector results.
            prob: If input number is probability.
        """
        super().__init__(dtype, force_input, prob, discretization)
        self._size = size
        self._element_col_name_template = element_col_name_template
        self._is_vector = is_vector

    @property
    def size(self):
        return self._size

    @property
    def is_vector(self):
        return self._is_vector

    def feature_name_at(self, position: int) -> str:
        """
        produces a name for feature on ``position`` in the vector

        Args:
            position: position in the vector in range [0 .. size]

        Returns:
            feature name

        """
        assert 0 <= position < self.size

        if isinstance(self._element_col_name_template, str):
            return self._element_col_name_template.format(position)

        return self._element_col_name_template[position]



"""Base classes for MLPipeline."""
import uuid
import warnings
from copy import copy
from typing import List, cast, Sequence, Union, Tuple, Optional

from pyspark.ml import Transformer, PipelineModel

from ..base import OutputFeaturesAndRoles
from ..features.base import SparkFeaturesPipeline, SparkEmptyFeaturePipeline
from ...dataset.base import LAMLDataset, SparkDataset
from ...ml_algo.base import SparkTabularMLAlgo
from ...transformers.base import ColumnsSelectorTransformer
from ...utils import Cacher
from ...validation.base import SparkBaseTrainValidIterator
from lightautoml.ml_algo.tuning.base import ParamsTuner
from lightautoml.ml_algo.utils import tune_and_fit_predict
from lightautoml.pipelines.ml.base import MLPipeline as LAMAMLPipeline
from lightautoml.pipelines.selection.base import SelectionPipeline


class SparkMLPipeline(LAMAMLPipeline, OutputFeaturesAndRoles):
    """Spark version of :class:`~lightautoml.pipelines.ml.base.MLPipeline`. Single ML pipeline.

    Merge together stage of building ML model
    (every step, excluding model training, is optional):

        - Pre selection: select features from input data.
          Performed by
          :class:`~lightautoml.pipelines.selection.base.SelectionPipeline`.
        - Features generation: build new features from selected.
          Performed by
          :class:`~sparklightautoml.pipelines.features.base.SparkFeaturesPipeline`.
        - Post selection: One more selection step - from created features.
          Performed by
          :class:`~lightautoml.pipelines.selection.base.SelectionPipeline`.
        - Hyperparams optimization for one or multiple ML models.
          Performed by
          :class:`~lightautoml.ml_algo.tuning.base.ParamsTuner`.
        - Train one or multiple ML models:
          Performed by :class:`~sparklightautoml.ml_algo.base.SparkTabularMLAlgo`.
          This step is the only required for at least 1 model.

    """

    def __init__(
        self,
        cacher_key: str,
        ml_algos: Sequence[Union[SparkTabularMLAlgo, Tuple[SparkTabularMLAlgo, ParamsTuner]]],
        force_calc: Union[bool, Sequence[bool]] = True,
        pre_selection: Optional[SelectionPipeline] = None,
        features_pipeline: Optional[SparkFeaturesPipeline] = None,
        post_selection: Optional[SelectionPipeline] = None,
        name: Optional[str] = None,
    ):
        if features_pipeline is None:
            features_pipeline = SparkEmptyFeaturePipeline()

        super().__init__(ml_algos, force_calc, pre_selection, features_pipeline, post_selection)

        self._cacher_key = cacher_key
        self._output_features = None
        self._output_roles = None
        self._transformer: Optional[Transformer] = None
        self._name = name if name else str(uuid.uuid4())[:5]
        self.ml_algos: List[SparkTabularMLAlgo] = []

    @property
    def name(self) -> str:
        return self._name

    @property
    def transformer(self):
        assert self._transformer is not None, f"{type(self)} seems to be not fitted"

        return self._transformer

    def fit_predict(self, train_valid: SparkBaseTrainValidIterator) -> LAMLDataset:
        """Fit on train/valid iterator and transform on validation part.

        Args:
            train_valid: Dataset iterator.

        Returns:
            Dataset with predictions of all models.

        """

        # train and apply pre selection
        input_roles = copy(train_valid.input_roles)
        full_train_roles = copy(train_valid.train.roles)

        train_valid = train_valid.apply_selector(self.pre_selection)

        # apply features pipeline
        train_valid = train_valid.apply_feature_pipeline(self.features_pipeline)
        fp = cast(SparkFeaturesPipeline, self.features_pipeline)

        # train and apply post selection
        train_valid = train_valid.apply_selector(self.post_selection)

        preds: Optional[SparkDataset] = None
        for ml_algo, param_tuner, force_calc in zip(self._ml_algos, self.params_tuners, self.force_calc):
            ml_algo = cast(SparkTabularMLAlgo, ml_algo)
            ml_algo, curr_preds = tune_and_fit_predict(ml_algo, param_tuner, train_valid, force_calc)
            if ml_algo is not None:
                self.ml_algos.append(ml_algo)
                preds = curr_preds
                train_valid.train = preds
            else:
                warnings.warn(
                    "Current ml_algo has not been trained by some reason. " "Check logs for more details.",
                    RuntimeWarning,
                )

        assert (
            len(self.ml_algos) > 0
        ), "Pipeline finished with 0 models for some reason.\nProbably one or more models failed"

        del self._ml_algos

        self._output_roles = {ml_algo.prediction_feature: ml_algo.prediction_role for ml_algo in self.ml_algos}

        # all out roles for the output dataset
        out_roles = copy(self._output_roles)
        # we need also add roles for predictions of previous pipe in this layer
        # because they are not part of either output roles of this pipe
        # (pipes work only with input data to the layer itself, e.g. independently)
        # nor input roles of train_valid iterator
        # (for each pipe iterator represent only input columns to the layer,
        # not outputs of other ml pipes in the layer)
        out_roles.update(full_train_roles)
        # we also need update our out_roles with input_roles to replace roles of input of the layer
        # in case they were changed by SparkChangeRolesTransformer
        out_roles.update(input_roles)

        select_transformer = ColumnsSelectorTransformer(
            input_cols=[SparkDataset.ID_COLUMN, *list(out_roles.keys())],
            optional_cols=[train_valid.train.target_column] if train_valid.train.target_column else [],
        )
        ml_algo_transformers = PipelineModel(stages=[ml_algo.transformer for ml_algo in self.ml_algos])
        self._transformer = PipelineModel(stages=[fp.transformer, ml_algo_transformers, select_transformer])

        val_preds = [preds.data]
        val_preds_df = train_valid.combine_val_preds(val_preds, include_train=False)
        val_preds_df = val_preds_df.select(
            SparkDataset.ID_COLUMN,
            train_valid.train.target_column,
            train_valid.train.folds_column,
            *list(out_roles.keys()),
        )

        cacher = Cacher(key=self._cacher_key)
        cacher.fit(val_preds_df)
        val_preds_df = cacher.dataset
        val_preds_ds = train_valid.train.empty()
        val_preds_ds.set_data(val_preds_df, list(out_roles.keys()), out_roles)

        return val_preds_ds

    def predict(self, dataset: SparkDataset) -> SparkDataset:
        """Predict on new dataset.

        Args:
            dataset: Dataset used for prediction.

        Returns:
            Dataset with predictions of all trained models.

        """
        out_sdf = self.transformer.transform(dataset.data)

        out_roles = copy(dataset.roles)
        out_roles.update(self.output_roles)

        out_ds = dataset.empty()
        out_ds.set_data(out_sdf, list(out_roles.keys()), out_roles)

        return out_ds



from copy import deepcopy
from typing import Optional, Union, Set

import numpy as np

from lightautoml.dataset.roles import CategoryRole, NumericRole
from lightautoml.pipelines.selection.base import ImportanceEstimator
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.pipelines.features.base import SparkFeaturesPipeline, SparkTabularDataFeatures
from sparklightautoml.transformers.base import (
    SparkChangeRolesTransformer,
    SparkUnionTransformer,
    SparkSequentialTransformer,
    SparkEstOrTrans,
)
from sparklightautoml.transformers.categorical import SparkOrdinalEncoderEstimator
from sparklightautoml.transformers.datetime import SparkTimeToNumTransformer


class SparkLGBSimpleFeatures(SparkFeaturesPipeline, SparkTabularDataFeatures):
    """Creates simple pipeline for tree based models.

    Simple but is ok for select features.
    Numeric stay as is, Datetime transforms to numeric.
    Categorical label encoding.
    Maps input to output features exactly one-to-one.

    """

    def __init__(self, cacher_key: str = "default_cacher"):
        super().__init__(cacher_key)

    def _get_input_features(self) -> Set[str]:
        return set(self.input_features)

    def create_pipeline(self, train: SparkDataset) -> Union[SparkUnionTransformer, SparkSequentialTransformer]:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        transformers_list = []

        # process categories
        categories = self._cols_by_role(train, "Category")
        if len(categories) > 0:
            roles = {f: train.roles[f] for f in categories}
            cat_processing = SparkOrdinalEncoderEstimator(
                input_cols=categories, input_roles=roles, subs=None, random_state=42
            )
            transformers_list.append(cat_processing)

        # process datetimes
        datetimes = self._cols_by_role(train, "Datetime")
        if len(datetimes) > 0:
            roles = {f: train.roles[f] for f in datetimes}
            dt_processing = SparkTimeToNumTransformer(input_cols=datetimes, input_roles=roles)
            transformers_list.append(dt_processing)

        transformers_list.append(self.get_numeric_data(train))

        union_all = SparkUnionTransformer([x for x in transformers_list if x is not None])

        return union_all


class SparkLGBAdvancedPipeline(SparkFeaturesPipeline, SparkTabularDataFeatures):
    """Create advanced pipeline for trees based models.

    Includes:

        - Different cats and numbers handling according to role params.
        - Dates handling - extracting seasons and create datediffs.
        - Create categorical intersections.

    """

    def __init__(
        self,
        feats_imp: Optional[ImportanceEstimator] = None,
        top_intersections: int = 5,
        max_intersection_depth: int = 3,
        subsample: Optional[Union[int, float]] = None,
        multiclass_te_co: int = 3,
        auto_unique_co: int = 10,
        output_categories: bool = False,
        cacher_key: str = "default_cacher",
        **kwargs
    ):
        """

        Args:
            feats_imp: Features importances mapping.
            top_intersections: Max number of categories
              to generate intersections.
            max_intersection_depth: Max depth of cat intersection.
            subsample: Subsample to calc data statistics.
            multiclass_te_co: Cutoff if use target encoding in cat
              handling on multiclass task if number of classes is high.
            auto_unique_co: Switch to target encoding if high cardinality.

        """
        super().__init__(
            cacher_key=cacher_key,
            multiclass_te_co=multiclass_te_co,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=subsample,
            feats_imp=feats_imp,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=False,
        )

    def _get_input_features(self) -> Set[str]:
        return set(self.input_features)

    def create_pipeline(self, train: SparkDataset) -> SparkEstOrTrans:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """

        features = train.features
        roles = deepcopy(train.roles)
        transformer_list = []

        target_encoder = self.get_target_encoder(train)

        output_category_role = (
            CategoryRole(np.float32, label_encoded=True) if self.output_categories else NumericRole(np.float32)
        )

        # handle categorical feats
        # split categories by handling type. This pipe use 3 encodings - freq/label/target/ordinal
        # 1 - separate freqs. It does not need label encoding
        stage = self.get_freq_encoding(train)
        transformer_list.append(stage)

        # 2 - check different target encoding parts and split (ohe is the same as auto - no ohe in gbm)
        auto = self._cols_by_role(train, "Category", encoding_type="auto") + self._cols_by_role(
            train, "Category", encoding_type="ohe"
        )

        if self.output_categories:
            le = (
                auto
                + self._cols_by_role(train, "Category", encoding_type="oof")
                + self._cols_by_role(train, "Category", encoding_type="int")
            )
            te = []
            ordinal = None

        else:
            le = self._cols_by_role(train, "Category", encoding_type="int")
            ordinal = self._cols_by_role(train, "Category", ordinal=True)

            if target_encoder is not None:
                te = self._cols_by_role(train, "Category", encoding_type="oof")
                # split auto categories by unique values cnt
                un_values = self.get_uniques_cnt(train, auto)
                te = te + [x for x in un_values.index if un_values[x] > self.auto_unique_co]
                ordinal = ordinal + list(set(auto) - set(te))

            else:
                te = []
                ordinal = ordinal + auto + self._cols_by_role(train, "Category", encoding_type="oof")

            ordinal = sorted(list(set(ordinal)))

        # get label encoded categories
        le_part = self.get_categorical_raw(train, le)
        if le_part is not None:
            # le_part = SequentialTransformer([le_part, ChangeRoles(output_category_role)])
            change_roles_stage = SparkChangeRolesTransformer(
                input_cols=le_part.getOutputCols(), input_roles=le_part.getOutputRoles(), role=output_category_role
            )
            le_part = SparkSequentialTransformer([le_part, change_roles_stage])
            transformer_list.append(le_part)

        # get target encoded part
        te_part = self.get_categorical_raw(train, te)
        if te_part is not None:
            target_encoder_stage = target_encoder(
                input_cols=te_part.getOutputCols(),
                input_roles=te_part.getOutputRoles(),
                task_name=train.task.name,
                folds_column=train.folds_column,
                target_column=train.target_column,
                do_replace_columns=True,
            )
            te_part = SparkSequentialTransformer([te_part, target_encoder_stage])
            transformer_list.append(te_part)

        # get intersection of top categories
        intersections = self.get_categorical_intersections(train)
        if intersections is not None:
            if target_encoder is not None:
                target_encoder_stage = target_encoder(
                    input_cols=intersections.getOutputCols(),
                    input_roles=intersections.getOutputRoles(),
                    task_name=train.task.name,
                    folds_column=train.folds_column,
                    target_column=train.target_column,
                    do_replace_columns=True,
                )
                ints_part = SparkSequentialTransformer([intersections, target_encoder_stage])
            else:
                change_roles_stage = SparkChangeRolesTransformer(
                    input_cols=intersections.getOutputCols(),
                    input_roles=intersections.getOutputRoles(),
                    role=output_category_role,
                )
                ints_part = SparkSequentialTransformer([intersections, change_roles_stage])

            transformer_list.append(ints_part)

        transformer_list.append(self.get_numeric_data(train))
        transformer_list.append(self.get_ordinal_encoding(train, ordinal))
        transformer_list.append(self.get_datetime_diffs(train))
        transformer_list.append(self.get_datetime_seasons(train, NumericRole(np.float32)))

        # final pipeline
        union_all = SparkUnionTransformer([x for x in transformer_list if x is not None])

        return union_all



import logging
from typing import Tuple, cast, List, Optional

import numpy as np
from pyspark.ml import PipelineModel, Transformer
from pyspark.ml.functions import vector_to_array
from pyspark.ml.param import Params
from pyspark.ml.param.shared import HasInputCols, HasOutputCol, Param
from pyspark.ml.util import DefaultParamsWritable, DefaultParamsReadable
from pyspark.sql import functions as F
from pyspark.sql.types import IntegerType

from sparklightautoml.spark_functions import array_to_vector, transform, aggregate

from lightautoml.dataset.roles import NumericRole, ColumnRole
from lightautoml.ml_algo.base import MLAlgo
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.pipelines.base import InputFeaturesAndRoles
from sparklightautoml.utils import Cacher, SparkDataFrame
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from lightautoml.utils.timer import TaskTimer

logger = logging.getLogger(__name__)

SparkMLModel = PipelineModel


class SparkTabularMLAlgo(MLAlgo, InputFeaturesAndRoles):
    """Machine learning algorithms that accepts numpy arrays as input."""

    _name: str = "SparkTabularMLAlgo"
    _default_validation_col_name: str = SparkBaseTrainValidIterator.TRAIN_VAL_COLUMN

    def __init__(
        self,
        cacher_key: str,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = {},
    ):
        super().__init__(default_params, freeze_defaults, timer, optimization_search_space)
        self._cacher_key = cacher_key
        self.n_classes: Optional[int] = None
        # names of columns that should contain predictions of individual models
        self._models_prediction_columns: Optional[List[str]] = None
        self._transformer: Optional[Transformer] = None

        # self._prediction_col = f"prediction_{self._name}"
        self._prediction_role = None

    @property
    def prediction_feature(self) -> str:
        # return self._prediction_col
        return f"prediction_{self._name}"


    @property
    def prediction_role(self) -> ColumnRole:
        return self._prediction_role

    @property
    def validation_column(self) -> str:
        return self._default_validation_col_name

    @property
    def transformer(self) -> Transformer:
        """Returns Spark MLlib Transformer.
        Represents a Transformer with fitted models."""

        assert self._transformer is not None, "Pipeline is not fitted!"

        return self._transformer

    def fit_predict(self, train_valid_iterator: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """

        logger.info(f"Input columns for MLALgo: {sorted(train_valid_iterator.input_features)}")
        logger.info(f"Train size for MLAlgo: {train_valid_iterator.train.data.count()}")

        # log_data(f"spark_fit_predict_{type(self).__name__}", {"train": train_valid_iterator.train.to_pandas()})

        assert self.is_fitted is False, "Algo is already fitted"
        # init params on input if no params was set before
        if self._params is None:
            self.params = self.init_params_on_input(train_valid_iterator)

        iterator_len = len(train_valid_iterator)
        if iterator_len > 1:
            logger.info("Start fitting \x1b[1m{}\x1b[0m ...".format(self._name))
            logger.debug(f"Training params: {self.params}")

        # save features names
        self._features = train_valid_iterator.features
        # get metric and loss if None
        self.task = train_valid_iterator.train.task

        valid_ds = cast(SparkDataset, train_valid_iterator.get_validation_data())

        # spark
        prob = train_valid_iterator.train.task.name in ["binary", "multiclass"]
        outp_dim = 1
        if self.task.name == "multiclass":
            outp_dim = valid_ds.data.select(F.max(valid_ds.target_column).alias("max")).first()
            outp_dim = outp_dim["max"] + 1
            self._prediction_role = NumericVectorOrArrayRole(
                outp_dim, f"{self.prediction_feature}" + "_{}", np.float32, force_input=True, prob=True
            )
        elif self.task.name == "binary":
            outp_dim = 2
            self._prediction_role = NumericVectorOrArrayRole(
                outp_dim, f"{self.prediction_feature}" + "_{}", np.float32, force_input=True, prob=True
            )
        else:
            self._prediction_role = NumericRole(np.float32, force_input=True, prob=False)

        self.n_classes = outp_dim

        preds_dfs: List[SparkDataFrame] = []

        pred_col_prefix = self._predict_feature_name()

        self._models_prediction_columns = []
        for n, (full, train, valid) in enumerate(train_valid_iterator):
            if iterator_len > 1:
                logger.info2(
                    "===== Start working with \x1b[1mfold {}\x1b[0m for \x1b[1m{}\x1b[0m =====".format(n, self._name)
                )
            self.timer.set_control_point()

            model_prediction_col = f"{pred_col_prefix}_{n}"
            model, val_pred, _ = self.fit_predict_single_fold(model_prediction_col, full, train, valid)

            self._models_prediction_columns.append(model_prediction_col)
            self.models.append(model)
            preds_dfs.append(val_pred)

            self.timer.write_run_info()

            if (n + 1) != len(train_valid_iterator):
                # split into separate cases because timeout checking affects parent pipeline timer
                if self.timer.time_limit_exceeded():
                    logger.info("Time limit exceeded after calculating fold {0}\n".format(n))
                    break

        # combining results for different folds
        # 1. folds - union
        # 2. dummy - nothing
        # 3. holdout - nothing
        # 4. custom - union + groupby
        # neutral_element = (
        #     array_to_vector(F.array(*[F.lit(float('nan')) for _ in range(self.n_classes)]))
        #     if self.task.name in ["binary", "multiclass"]
        #     else F.lit(float('nan'))
        # )

        neutral_element = None

        preds_dfs = [
            df.select(
                "*", *[F.lit(neutral_element).alias(c) for c in self._models_prediction_columns if c not in df.columns]
            )
            for df in preds_dfs
        ]
        full_preds_df = train_valid_iterator.combine_val_preds(preds_dfs, include_train=False)
        full_preds_df = self._build_averaging_transformer().transform(full_preds_df)

        cacher = Cacher(key=self._cacher_key)
        cacher.fit(full_preds_df)
        full_preds_df = cacher.dataset

        # create Spark MLlib Transformer and save to property var
        self._transformer = self._build_transformer()

        pred_ds = self._set_prediction(valid_ds.empty(), full_preds_df)

        if iterator_len > 1:
            single_pred_ds = self._make_single_prediction_dataset(pred_ds)
            logger.info(
                f"Fitting \x1b[1m{self._name}\x1b[0m finished. score = \x1b[1m{self.score(single_pred_ds)}\x1b[0m"
            )

        if iterator_len > 1 or "Tuned" not in self._name:
            logger.info("\x1b[1m{}\x1b[0m fitting and predicting completed".format(self._name))

        return pred_ds

    def fit_predict_single_fold(
        self, fold_prediction_column: str, full: SparkDataset, train: SparkDataset, valid: SparkDataset
    ) -> Tuple[SparkMLModel, SparkDataFrame, str]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            fold_prediction_column: column name for predictions made for this fold
            full: Full dataset that include train and valid parts and a bool column that delimits records
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Target predictions for valid dataset.

        """
        raise NotImplementedError

    def predict_single_fold(self, model: SparkMLModel, dataset: SparkDataset) -> SparkDataFrame:
        raise NotImplementedError("Not supported for Spark. Use transformer property instead ")
        pass

    def predict(self, dataset: SparkDataset) -> SparkDataset:
        sdf = self.transformer.transform(dataset.data)

        ds = dataset.empty()
        ds.set_data(sdf, [self.prediction_feature], self.prediction_role)

        return ds

    @staticmethod
    def _get_predict_column(model: SparkMLModel) -> str:
        try:
            return model.getPredictionCol()
        except AttributeError:
            if isinstance(model, PipelineModel):
                return model.stages[-1].getPredictionCol()

            raise TypeError("Unknown model type! Unable ro retrieve prediction column")

    def _predict_feature_name(self):
        return f"{self._name}_prediction"

    def _set_prediction(self, dataset: SparkDataset, preds: SparkDataFrame) -> SparkDataset:
        """Insert predictions to dataset with. Inplace transformation.

        Args:
            dataset: Dataset to transform.
            preds: A spark dataframe  with predicted values.

        Returns:
            Transformed dataset.

        """

        prob = self.task.name in ["binary", "multiclass"]

        if self.task.name in ["binary", "multiclass"]:
            role = NumericVectorOrArrayRole(
                size=self.n_classes,
                element_col_name_template=self._predict_feature_name() + "_{}",
                dtype=np.float32,
                force_input=True,
                prob=prob,
            )
        else:
            role = NumericRole(dtype=np.float32, force_input=True, prob=prob)

        output: SparkDataset = dataset.empty()
        output.set_data(preds, [self._predict_feature_name()], role)

        return output

    def _build_transformer(self) -> Transformer:
        raise NotImplementedError()

    def _build_averaging_transformer(self) -> Transformer:
        raise NotImplementedError()

    def _make_single_prediction_dataset(self, dataset: SparkDataset) -> SparkDataset:
        preds = dataset.data.select(SparkDataset.ID_COLUMN, dataset.target_column, self.prediction_feature)
        roles = {self.prediction_feature: dataset.roles[self.prediction_feature]}

        output: SparkDataset = dataset.empty()
        output.set_data(preds, preds.columns, roles)

        return output


class AveragingTransformer(Transformer, HasInputCols, HasOutputCol, DefaultParamsWritable, DefaultParamsReadable):
    """
    Transformer that gets one or more columns and produce column with average values.
    """

    taskName = Param(Params._dummy(), "taskName", "task name")
    removeCols = Param(Params._dummy(), "removeCols", "cols to remove")
    convertToArrayFirst = Param(Params._dummy(), "convertToArrayFirst", "convert to array first")
    weights = Param(Params._dummy(), "weights", "weights")
    dimNum = Param(Params._dummy(), "dimNum", "dim num")

    def __init__(
        self,
        task_name: str = None,
        input_cols: List[str] = [],
        output_col: str = "averaged_values",
        remove_cols: Optional[List[str]] = None,
        convert_to_array_first: bool = False,
        weights: Optional[List[int]] = None,
        dim_num: int = 1,
    ):
        """
        Args:
            task_name (str, optional): Task name: "binary", "multiclass" or "reg".
            input_cols (List[str], optional): List of input columns.
            output_col (str, optional): Output column name. Defaults to "averaged_values".
            remove_cols (Optional[List[str]], optional): Columns need to remove. Defaults to None.
            convert_to_array_first (bool, optional): If `True` then will be convert input vectors to arrays. Defaults to False.
            weights (Optional[List[int]], optional): List of weights to scaling output values. Defaults to None.
            dim_num (int, optional): Dimension of input columns. Defaults to 1.
        """
        super().__init__()
        self.set(self.taskName, task_name)
        self.set(self.inputCols, input_cols)
        self.set(self.outputCol, output_col)
        if not remove_cols:
            remove_cols = []
        self.set(self.removeCols, remove_cols)
        self.set(self.convertToArrayFirst, convert_to_array_first)
        if weights is None:
            weights = [1.0 for _ in input_cols]

        assert len(input_cols) == len(weights)

        self.set(self.weights, weights)
        self.set(self.dimNum, dim_num)

    def getTaskName(self) -> str:
        return self.getOrDefault(self.taskName)

    def getRemoveCols(self) -> List[str]:
        return self.getOrDefault(self.removeCols)

    def getConvertToArrayFirst(self) -> bool:
        return self.getOrDefault(self.convertToArrayFirst)

    def getWeights(self) -> List[int]:
        return self.getOrDefault(self.weights)

    def getDimNum(self) -> int:
        return self.getOrDefault(self.dimNum)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.info(f"In transformer {type(self)}. Columns: {sorted(dataset.columns)}")

        pred_cols = self.getInputCols()
        dim_size = self.getInputCols()
        weights = {c: w for w, c in zip(self.getWeights(), pred_cols)}
        non_null_count_col = F.lit(len(pred_cols)) - sum(F.isnull(c).astype(IntegerType()) for c in pred_cols)

        if self.getTaskName() in ["binary", "multiclass"]:

            def convert_column(c):
                return vector_to_array(c).alias(c) if self.getConvertToArrayFirst() else F.col(c)

            normalized_cols = [
                F.when(F.isnull(c), F.array(*[F.lit(0.0) for _ in range(self.getDimNum())]))
                .otherwise(convert_column(c))
                .alias(c)
                for c in pred_cols
            ]
            arr_fields_summ = F.transform(
                F.arrays_zip(*normalized_cols),
                lambda x: F.aggregate(
                    F.array(*[x[c] * F.lit(weights[c]) for c in pred_cols]), F.lit(0.0), lambda acc, x: acc + x
                )
                / non_null_count_col,
            )

            out_col = array_to_vector(arr_fields_summ) if self.getConvertToArrayFirst() else arr_fields_summ
        else:
            scalar_fields_summ = (
                F.aggregate(
                    F.array(*[F.col(c) * F.lit(weights[c]) for c in pred_cols]),
                    F.lit(0.0),
                    lambda acc, x: acc + F.when(F.isnull(x), F.lit(0.0)).otherwise(x),
                )
                / non_null_count_col
            )

            out_col = scalar_fields_summ

        cols_to_remove = set(self.getRemoveCols())
        cols_to_select = [c for c in dataset.columns if c not in cols_to_remove]
        out_df = dataset.select(*cols_to_select, out_col.alias(self.getOutputCol()))

        logger.info(f"In the end of transformer {type(self)}. Columns: {sorted(dataset.columns)}")

        return out_df



import logging
import multiprocessing
import warnings
from copy import copy
from typing import Dict, Optional, Tuple, Union, cast

import pandas as pd
import pyspark.sql.functions as F
from pandas import Series
from pyspark.ml import Transformer, PipelineModel
from pyspark.ml.util import MLWritable, MLReadable, MLWriter
from pyspark.ml.feature import VectorAssembler
from synapse.ml.lightgbm import (
    LightGBMClassifier,
    LightGBMRegressor,
    LightGBMRegressionModel,
    LightGBMClassificationModel,
)
from synapse.ml.onnx import ONNXModel
import lightgbm as lgb
from lightgbm import Booster, LGBMClassifier


from lightautoml.ml_algo.tuning.base import Distribution, SearchSpace
from lightautoml.pipelines.selection.base import ImportanceEstimator
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.ml_algo.base import SparkTabularMLAlgo, SparkMLModel, AveragingTransformer
from sparklightautoml.mlwriters import (
    LightGBMModelWrapperMLReader,
    LightGBMModelWrapperMLWriter,
    ONNXModelWrapperMLReader,
    ONNXModelWrapperMLWriter,
)
from sparklightautoml.transformers.base import (
    DropColumnsTransformer,
    PredictionColsTransformer,
    ProbabilityColsTransformer,
)
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from lightautoml.utils.timer import TaskTimer
from lightautoml.validation.base import TrainValidIterator

logger = logging.getLogger(__name__)


class LightGBMModelWrapper(Transformer, MLWritable, MLReadable):
    """Simple wrapper for `synapse.ml.lightgbm.[LightGBMRegressionModel|LightGBMClassificationModel]` to fix issue with loading model from saved composite pipeline.

    For more details see: https://github.com/microsoft/SynapseML/issues/614.
    """

    def __init__(self, model: Union[LightGBMRegressionModel, LightGBMClassificationModel] = None) -> None:
        super().__init__()
        self.model = model

    def write(self) -> MLWriter:
        return LightGBMModelWrapperMLWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return LightGBMModelWrapperMLReader()

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        return self.model.transform(dataset)


class ONNXModelWrapper(Transformer, MLWritable, MLReadable):
    """Simple wrapper for `ONNXModel` to fix issue with loading model from saved composite pipeline.

    For more details see: https://github.com/microsoft/SynapseML/issues/614.
    """

    def __init__(self, model: ONNXModel = None) -> None:
        super().__init__()
        self.model = model

    def write(self) -> MLWriter:
        return ONNXModelWrapperMLWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return ONNXModelWrapperMLReader()

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        return self.model.transform(dataset)


class SparkBoostLGBM(SparkTabularMLAlgo, ImportanceEstimator):
    """Gradient boosting on decision trees from LightGBM library.

    default_params: All available parameters listed in synapse.ml documentation:

        - https://mmlspark.blob.core.windows.net/docs/0.9.5/pyspark/synapse.ml.lightgbm.html#module-synapse.ml.lightgbm.LightGBMClassifier
        - https://mmlspark.blob.core.windows.net/docs/0.9.5/pyspark/synapse.ml.lightgbm.html#module-synapse.ml.lightgbm.LightGBMRegressor

    freeze_defaults:

        - ``True`` :  params may be rewritten depending on dataset.
        - ``False``:  params may be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "LightGBM"

    _default_params = {
        "learningRate": 0.05,
        "numLeaves": 128,
        "featureFraction": 0.7,
        "baggingFraction": 0.7,
        "baggingFreq": 1,
        "maxDepth": -1,
        "minGainToSplit": 0.0,
        "maxBin": 255,
        "minDataInLeaf": 5,
        # e.g. num trees
        "numIterations": 3000,
        "earlyStoppingRound": 50,
        # for regression
        "alpha": 1.0,
        "lambdaL1": 0.0,
        "lambdaL2": 0.0,
    }

    # mapping between metric name defined via SparkTask
    # and metric names supported by LightGBM
    _metric2lgbm = {
        "binary": {"auc": "auc", "aupr": "areaUnderPR"},
        "reg": {
            "r2": "rmse",
            "mse": "mse",
            "mae": "mae",
        },
        "multiclass": {"crossentropy": "cross_entropy"},
    }

    def __init__(
        self,
        cacher_key: str,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = {},
        use_single_dataset_mode: bool = True,
        max_validation_size: int = 10_000,
        chunk_size: int = 4_000_000,
        convert_to_onnx: bool = False,
        mini_batch_size: int = 5000,
        seed: int = 42,
    ):
        SparkTabularMLAlgo.__init__(self, cacher_key, default_params, freeze_defaults, timer, optimization_search_space)
        self._probability_col_name = "probability"
        self._prediction_col_name = "prediction"
        self._raw_prediction_col_name = "raw_prediction"
        self._assembler = None
        self._drop_cols_transformer = None
        self._use_single_dataset_mode = use_single_dataset_mode
        self._max_validation_size = max_validation_size
        self._seed = seed
        self._models_feature_impotances = []
        self._chunk_size = chunk_size
        self._convert_to_onnx = convert_to_onnx
        self._mini_batch_size = mini_batch_size

    def _infer_params(self) -> Tuple[dict, int]:
        """Infer all parameters in lightgbm format.

        Returns:
            Tuple (params, num_trees, early_stopping_rounds, verbose_eval, fobj, feval).
            About parameters: https://lightgbm.readthedocs.io/en/latest/_modules/lightgbm/engine.html

        """
        assert self.task is not None

        task = self.task.name

        params = copy(self.params)

        if "isUnbalance" in params:
            params["isUnbalance"] = True if params["isUnbalance"] == 1 else False

        verbose_eval = 1

        if task == "reg":
            params["objective"] = "regression"
            params["metric"] = self._metric2lgbm[task].get(self.task.metric_name, None)
        elif task == "binary":
            params["objective"] = "binary"
            params["metric"] = self._metric2lgbm[task].get(self.task.metric_name, None)
        elif task == "multiclass":
            params["objective"] = "multiclass"
            params["metric"] = "multiclass"
        else:
            raise ValueError(f"Unsupported task type: {task}")

        if task != "reg":
            if "alpha" in params:
                del params["alpha"]
            if "lambdaL1" in params:
                del params["lambdaL1"]
            if "lambdaL2" in params:
                del params["lambdaL2"]

        params = {**params}

        return params, verbose_eval

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        self.task = train_valid_iterator.train.task

        sds = cast(SparkDataset, train_valid_iterator.train)
        rows_num = sds.data.count()
        task = train_valid_iterator.train.task.name

        suggested_params = copy(self.default_params)

        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params

        if task == "reg":
            suggested_params = {
                "learningRate": 0.05,
                "numLeaves": 32,
                "featureFraction": 0.9,
                "baggingFraction": 0.9,
            }

        if rows_num <= 10000:
            init_lr = 0.01
            ntrees = 3000
            es = 200

        elif rows_num <= 20000:
            init_lr = 0.02
            ntrees = 3000
            es = 200

        elif rows_num <= 100000:
            init_lr = 0.03
            ntrees = 2000
            es = 200
        elif rows_num <= 300000:
            init_lr = 0.04
            ntrees = 2000
            es = 100
        else:
            init_lr = 0.05
            ntrees = 2000
            es = 100

        if rows_num > 300000:
            suggested_params["numLeaves"] = 128 if task == "reg" else 244
        elif rows_num > 100000:
            suggested_params["numLeaves"] = 64 if task == "reg" else 128
        elif rows_num > 50000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 64
            # params['reg_alpha'] = 1 if task == 'reg' else 0.5
        elif rows_num > 20000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 32
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.0
        elif rows_num > 10000:
            suggested_params["numLeaves"] = 32 if task == "reg" else 64
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.2
        elif rows_num > 5000:
            suggested_params["numLeaves"] = 24 if task == "reg" else 32
            suggested_params["alpha"] = 0.5 if task == "reg" else 0.5
        else:
            suggested_params["numLeaves"] = 16 if task == "reg" else 16
            suggested_params["alpha"] = 1 if task == "reg" else 1

        suggested_params["learningRate"] = init_lr
        suggested_params["numIterations"] = ntrees
        suggested_params["earlyStoppingRound"] = es

        if task != "reg":
            if "alpha" in suggested_params:
                del suggested_params["alpha"]

        return suggested_params

    def _get_default_search_spaces(self, suggested_params: Dict, estimated_n_trials: int) -> Dict:
        """Train on train dataset and predict on holdout dataset.

        Args:
            fold_prediction_column: column name for predictions made for this fold
            full: Full dataset that include train and valid parts and a bool column that delimits records
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Target predictions for valid dataset.

        """
        assert self.task is not None

        optimization_search_space = dict()

        optimization_search_space["featureFraction"] = SearchSpace(
            Distribution.UNIFORM,
            low=0.5,
            high=1.0,
        )

        optimization_search_space["numLeaves"] = SearchSpace(
            Distribution.INTUNIFORM,
            low=4,
            high=255,
        )

        if self.task.name == "binary" or self.task.name == "multiclass":
            optimization_search_space["isUnbalance"] = SearchSpace(Distribution.DISCRETEUNIFORM, low=0, high=1, q=1)

        if estimated_n_trials > 30:
            optimization_search_space["baggingFraction"] = SearchSpace(
                Distribution.UNIFORM,
                low=0.5,
                high=1.0,
            )

            optimization_search_space["minSumHessianInLeaf"] = SearchSpace(
                Distribution.LOGUNIFORM,
                low=1e-3,
                high=10.0,
            )

        if estimated_n_trials > 100:
            if self.task.name == "reg":
                optimization_search_space["alpha"] = SearchSpace(
                    Distribution.LOGUNIFORM,
                    low=1e-8,
                    high=10.0,
                )

            optimization_search_space["lambdaL1"] = SearchSpace(
                Distribution.LOGUNIFORM,
                low=1e-8,
                high=10.0,
            )

        return optimization_search_space

    def predict_single_fold(
        self, dataset: SparkDataset, model: Union[LightGBMRegressor, LightGBMClassifier]
    ) -> SparkDataFrame:

        temp_sdf = self._assembler.transform(dataset.data)

        pred = model.transform(temp_sdf)

        return pred

    def fit_predict_single_fold(
        self, fold_prediction_column: str, full: SparkDataset, train: SparkDataset, valid: SparkDataset
    ) -> Tuple[SparkMLModel, SparkDataFrame, str]:
        assert self.validation_column in full.data.columns, "Train should contain validation column"

        if self.task is None:
            self.task = full.task

        (params, verbose_eval) = self._infer_params()

        logger.info(f"Input cols for the vector assembler: {full.features}")
        logger.info(f"Running lgb with the following params: {params}")

        if self._assembler is None:
            self._assembler = VectorAssembler(
                inputCols=self.input_features, outputCol=f"{self._name}_vassembler_features", handleInvalid="keep"
            )

        LGBMBooster = LightGBMRegressor if full.task.name == "reg" else LightGBMClassifier

        if full.task.name in ["binary", "multiclass"]:
            params["rawPredictionCol"] = self._raw_prediction_col_name
            params["probabilityCol"] = fold_prediction_column
            params["predictionCol"] = self._prediction_col_name
            params["isUnbalance"] = True
        else:
            params["predictionCol"] = fold_prediction_column

        master_addr = train.spark_session.conf.get("spark.master")
        if master_addr.startswith("local"):
            cores_str = master_addr[len("local[") : -1]
            cores = int(cores_str) if cores_str != "*" else multiprocessing.cpu_count()
            params["numThreads"] = max(cores - 1, 1)
        else:
            params["numThreads"] = max(int(train.spark_session.conf.get("spark.executor.cores", "1")) - 1, 1)

        train_data = full.data
        valid_size = train_data.where(F.col(self.validation_column) == 1).count()
        max_val_size = self._max_validation_size
        if valid_size > max_val_size:
            warnings.warn(
                f"Maximum validation size for SparkBoostLGBM is exceeded: {valid_size} > {max_val_size}. "
                f"Reducing validation size down to maximum.",
                category=RuntimeWarning,
            )
            rest_cols = list(train_data.columns)
            rest_cols.remove(self.validation_column)

            replace_col = F.when(F.rand(self._seed) < max_val_size / valid_size, F.lit(True)).otherwise(F.lit(False))
            val_filter_cond = F.when(F.col(self.validation_column) == 1, replace_col).otherwise(F.lit(True))

            train_data = train_data.where(val_filter_cond)

        valid_data = valid.data

        lgbm = LGBMBooster(
            **params,
            featuresCol=self._assembler.getOutputCol(),
            labelCol=full.target_column,
            validationIndicatorCol=self.validation_column,
            verbosity=verbose_eval,
            useSingleDatasetMode=self._use_single_dataset_mode,
            isProvideTrainingMetric=True,
            chunkSize=self._chunk_size,
        )

        logger.info(f"Use single dataset mode: {lgbm.getUseSingleDatasetMode()}. NumThreads: {lgbm.getNumThreads()}")

        if full.task.name == "reg":
            lgbm.setAlpha(0.5).setLambdaL1(0.0).setLambdaL2(0.0)

        ml_model = lgbm.fit(self._assembler.transform(train_data))

        val_pred = ml_model.transform(self._assembler.transform(valid_data))
        val_pred = DropColumnsTransformer(
            remove_cols=[],
            optional_remove_cols=[self._prediction_col_name, self._probability_col_name, self._raw_prediction_col_name],
        ).transform(val_pred)

        self._models_feature_impotances.append(ml_model.getFeatureImportances(importance_type="gain"))

        if self._convert_to_onnx:
            logger.info("Model convert is started")
            booster_model_str = ml_model.getLightGBMBooster().modelStr().get()
            booster = lgb.Booster(model_str=booster_model_str)
            model_payload_ml = self._convertModel(booster, len(self.input_features))

            onnx_ml = ONNXModel().setModelPayload(model_payload_ml)

            if full.task.name == "reg":
                onnx_ml = (
                    onnx_ml.setDeviceType("CPU")
                    .setFeedDict({"input": f"{self._name}_vassembler_features"})
                    .setFetchDict({ml_model.getPredictionCol(): "variable"})
                    .setMiniBatchSize(self._mini_batch_size)
                )
            else:
                onnx_ml = (
                    onnx_ml.setDeviceType("CPU")
                    .setFeedDict({"input": f"{self._name}_vassembler_features"})
                    .setFetchDict({ml_model.getProbabilityCol(): "probabilities", ml_model.getPredictionCol(): "label"})
                    .setMiniBatchSize(self._mini_batch_size)
                )

            ml_model = onnx_ml
            logger.info("Model convert is ended")

        return ml_model, val_pred, fold_prediction_column

    def fit(self, train_valid: SparkBaseTrainValidIterator):
        logger.info("Starting LGBM fit")
        self.fit_predict(train_valid)
        logger.info("Finished LGBM fit")

    def get_features_score(self) -> Series:
        imp = 0
        for model_feature_impotances in self._models_feature_impotances:
            imp = imp + pd.Series(model_feature_impotances)

        imp = imp / len(self._models_feature_impotances)

        result = Series(list(imp), index=self.features).sort_values(ascending=False)
        return result

    @staticmethod
    def _convertModel(lgbm_model: Booster, input_size: int) -> bytes:
        from onnxmltools.convert import convert_lightgbm
        from onnxconverter_common.data_types import FloatTensorType

        initial_types = [("input", FloatTensorType([-1, input_size]))]
        onnx_model = convert_lightgbm(lgbm_model, initial_types=initial_types, target_opset=9)
        return onnx_model.SerializeToString()

    def _build_transformer(self) -> Transformer:
        avr = self._build_averaging_transformer()
        if self._convert_to_onnx:
            wrapped_models = [ONNXModelWrapper(m) for m in self.models]
        else:
            wrapped_models = [LightGBMModelWrapper(m) for m in self.models]
        models = [
            el
            for m in wrapped_models
            for el in [
                m,
                DropColumnsTransformer(
                    remove_cols=[],
                    optional_remove_cols=[
                        self._prediction_col_name,
                        self._probability_col_name,
                        self._raw_prediction_col_name,
                    ],
                ),
            ]
        ]
        if self._convert_to_onnx:
            if self.task.name in ["binary", "multiclass"]:
                models.append(
                    ProbabilityColsTransformer(
                        probability_сols=self._models_prediction_columns, num_classes=self.n_classes
                    )
                )
            else:
                models.append(PredictionColsTransformer(prediction_сols=self._models_prediction_columns))
        averaging_model = PipelineModel(stages=[self._assembler] + models + [avr])
        return averaging_model

    def _build_averaging_transformer(self) -> Transformer:
        avr = AveragingTransformer(
            self.task.name,
            input_cols=self._models_prediction_columns,
            output_col=self.prediction_feature,
            remove_cols=[self._assembler.getOutputCol()] + self._models_prediction_columns,
            convert_to_array_first=not (self.task.name == "reg"),
            dim_num=self.n_classes,
        )
        return avr

    def fit_predict(self, train_valid_iterator: SparkBaseTrainValidIterator) -> SparkDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        logger.info("Starting LGBM fit")
        self.timer.start()

        self.input_roles = train_valid_iterator.input_roles

        res = super().fit_predict(train_valid_iterator)

        logger.info("Finished LGBM fit")
        return res
